#!python
#
# OS2SRTM1.py
#
# Charles Macfarlane, May/June 2015
#
# First of five programs/scripts to convert OS Landform Panorama or Terrain 50 elevation data to SRTM format -
# the programs/scripts are:
#	OS2SRTM1.py		Begins by making for each possible SRTM cell a WGS84 coordinate line in a *.pt? file
#					Requires GeoidEval from GeographicLib to be on the system path: http://geographiclib.sourceforge.net/
# Either:
#	GridInQuest		External OS program translates the WSG84 coordinates from above into OSGB36 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/help-and-support/navigation-technology/os-net/grid-inquest.html
# Or:
#	OS2SRTM2.js		Translates the WSG84 coordinates from above into OSGB36 and appends them
#					Unlike GIQ, this script cannot convert geiod heights, and therefore the /G switch cannot be
#					used in the final script
#	OS2SRTM3.py		Looks up heights from Landform Panorama and Terrain 50 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/products/land-form-panorama.html
#					http://www.ordnancesurvey.co.uk/business-and-government/products/terrain-50.html
#	OS2SRTM4.py		Optionally uses a bitmap OSBitMap.bin to remove points corresponding to sea
#	OS2SRTM5.py		Creates SRTM tiles
#
# This first script makes in the current directory for each degree of latitude a *.pt1 or *.pt2
# height correction comma-seperated-value (CSV) file with line format ...
#	<SRTM TileName>	For example N55W006
#	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
#	<Tile Col>		Ditto
#	<Lat>			Latitude in decimal degrees corresponding to row above
#	<Lon>			Longitude in decimal degrees corresponding to col above
#	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
# ... for example for N59W003.pt1 ...
#	N59W003,0000,0000,59.0000000,-3.0000000,52.465
#
# This script creates a list of the SRTM sample points, then looks up and appends the geoid height for each
# using GeoidEval from the geographic library GeographicLib, which must be available on the system path.
#
# The next stage after running this script is ...
# EITHER:
#	Run the batch conversion utility of Grid InQuest to transform the last three fields above into OSGB36 and
#	append them to the lines in the existing file.  Do not worry that many lines representing points in the sea
#	will cause GIQ to append the message ...
#		Coordinate outside transformation boundary
# 	... as the next script will simply ignore such lines. The exact procedure for using GIQ is as follows:
#	1	Launch the Grid InQuest program
#	2	Choose File, Convert file(s), and then choose the appropriate file(s)
#	3	If you have not previously saved the conversion configuration, leave it on the default ...
#			<New Format>
#		... otherwise select your previously saved configuration, then click Next.
#	4	Click Next on the next page without changing anything, thereby starting import at row 1.
#	5	Click Next on the next page without changing anything, thereby choosing Delimited.
#	6	Click Next on the next page without changing anything, thereby choosing Comma as the delimiter.
#	7	Select Geodetic (Latitude, Longitude) and click Next
#	8	At the next page ...
#			<Rt-Click> column 4 and select Latitude
#			<Rt-Click> column 5 and select Longitude
#			<Rt-Click> column 6 and select Ellipsoidal Height
#		... and then click Next.
#	9	Select Projected (Eastings, Northings) and, before clicking Next, check that the remaining fields
#		are at the defaults:
#			OSGB 1936, Airy 1830, OSGB36 (OSTN02), United Kingdom.
#	10	Select Append the converted coordinates to the existing file(s), and then click Next.  Confusingly,
#		sometimes this setting is remembered between successive runs, sometimes not, so always check it.
#	11	If you haven't already saved the conversion configuration, save it giving it a suitable name -
#		I suggest OS2SRTM.  Then click Finish to start the conversion.
# OR:
#	Run OS2SRTM2.js to convert the Lat/Lon only and with less accuracy
#
# The *.pt1/*.pt2 data lines will then have the format ...
#	<SRTM TileName>	For example N55W006
#	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
#	<Tile Col>		Ditto
#	<Lat>			Latitude in decimal degrees corresponding to row above
#	<Lon>			Longitude in decimal degrees corresponding to col above
#	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
#	<East>			Easting in metres - translated from above (Lat,Lon,Ht) by Grid In Quest
#	<North>			Northing in metres - ditto
#	<EGM96 v OD>	Height in metres of the EGM96 geoid above Ordnance Datum - ditto
#	<OD Name>		Name of above Ordnance Datum
# ... for example for N59W003.pt1 ...
#	N59W003,0000,0000,59.0000000,-3.0000000,52.465,342647,1012974,0.874,Newlyn (Orkney Isles only)
#
# Then run the remaining scripts in the numerical order listed above, resuming with OS2SRTM3.py
#
# Command-line switches:
#	/A	Use alternative cell centering and correspondingly create *.pt2 file(s)
#		(*.pt1 = normal, *.pt2 = alternative - also see note 1)
#	/S	Do the six tiles covering the Shetlands only
#		(at latitude 59N, ignore corrections at or below SRTM3 row 540, SRTM1 row 1620 - also see note 2)
#	/T	Do a test run across the middle of the country
#		(see note 2)
#	/1	Create SRTM1 tiles, default is SRTM3
#		(see note 1)
#
# Notes:
#
#	1	The /A and /1 switches MUST be used consistently throughout all the scripts.
#	2	The /S and /T switches can be introduced at any stage, but MUST be used consistently thereafter. These
#		switches are mutually exclusive, if both are given, whichever is given second is ignored.
#	3	The /G switch in the final script is valid only if Grid InQuest is used at stage 2.
#	4	Debug output is often best redirected to file, because when output is voluminous the need to scroll it
#		slows program execution.
#
# ***********************************************************************
#
# Charles Macfarlane's Work is licensed under a Creative Commons Licence -
# Attribution-Non-Commercial-No Derivative Works 2.0 UK: England & Wales
# http://creativecommons.org/licenses/by-nc-nd/2.0/uk/
#
# For further details, please see:
# http://www.macfh.co.uk/JavaJive/JavaJive.html
#
# ***********************************************************************


import os
import sys
import re
import subprocess
import string

PtsPath		= ".\\"
SRTMPath	= ".\\..\\"
SRTMExt		= ".hgt"
ZipExt		= ".zip"
Pts1Ext		= ".pt1"
Pts2Ext		= ".pt2"
bakExt		= ".bak"
BIFileName	= "OS2SRTM-BI-N%d"
ShFileName	= "OS2SRTM-Shetland-N%d"
TstFileName	= "OS2SRTM-Test"
TileNameT	= "%s%02d%s%03d"
BIStartX	= -8
BIStartY	= 49
BIEndX		= 2
BIEndY		= 61
ShStartX	= -3
ShStartY	= 59
ShEndX		= 0
ShEndY		= 61
TstStartX	= -7
TstStartY	= 55
TstEndX		= 0
TstEndY		= 56
startX		= BIStartX
startY		= BIStartY
endX		= BIEndX
endY		= BIEndY
SRTM1Cols	= 3601
SRTM1Rows	= 3601
SRTM3Cols	= 1201
SRTM3Rows	= 1201
SRTMCols	= SRTM3Cols
SRTMRows	= SRTM3Rows

# Number of decimal places for the heights
Accuracy = 3
# Value flagging an invalid height
InvHeight	= -32768;

# Flags which cell alignment method to use:
#	False	Use SW corner of SRTM and OS Landform Panorama cells, adjust to middle of Terrain 50 cells
#	True	Adjust to middle of SRTM and OS Landform Panorama cells, use SW corner of Terrain 50 cells
Alternate	= False
# Geoid flag
Geoid		= True
# Shetland flag
Shetland	= False
# Test flag
Test		= False

# Main program entry point
def main():

	DataNameT	= BIFileName
	ptsExt		= Pts1Ext
	global	Alternate, Geoid, Shetland, Test, startX, startY, endX, endY, SRTMCols, SRTMRows

	for opt in sys.argv[1:]:
		if re.compile( "^[-/]G-$", re.IGNORECASE ).match( opt ):
			Geoid		= False
			print		"Skipping Geoid heights ..."
		if re.compile( "^[-/]A$", re.IGNORECASE ).match( opt ):
			Alternate	= True
			ptsExt		= Pts2Ext
			print		"Alternate cell alignment method ..."
		elif re.compile( "^[-/]S$", re.IGNORECASE ).match( opt ) and not Test:
			Shetland	= True
			startX		= ShStartX
			startY		= ShStartY
			endX		= ShEndX
			endY		= ShEndY
			DataNameT	= ShFileName
			print		"Shetland only ..."
		elif re.compile( "^[-/]T$", re.IGNORECASE ).match( opt ) and not Shetland:
			Test		= True
			startX		= TstStartX
			startY		= TstStartY
			endX		= TstEndX
			endY		= TstEndY
			DataNameT	= TstFileName
			print		"Testing only ..."
		elif re.compile( "^[-/]1$", re.IGNORECASE ).match( opt ):
			SRTMCols	= SRTM1Cols
			SRTMRows	= SRTM1Rows
			print		"SRTM 1 ArcSecond points ..."

	DataNameT	= os.path.join( os.path.dirname(__file__), DataNameT )
	DataIName	= os.path.join( os.path.dirname(__file__), "Input.dat" )
	DataOName	= os.path.join( os.path.dirname(__file__), "Output.dat" )

	for lat in range( startY, endY ):

		# Don't want an error - Test option does not require the substitution
		try:
			DataFName = DataNameT % ( lat )
		except:
			DataFName = DataNameT

		if os.access(DataFName + ptsExt, os.F_OK) and os.access(DataFName + bakExt, os.F_OK):
			os.remove( DataFName + bakExt )
		if os.access( DataFName + ptsExt, os.F_OK ):
			os.rename( DataFName + ptsExt, DataFName + bakExt )
		DataFName += ptsExt
		print "Processing file: %s" % ( DataFName )

		# Create the master points file and the next stage's input file
		if Geoid:
			Inp = open( DataIName, "w" )
		Outp = open( DataFName, "w" )

		for lon in range( startX, endX ):
			TileName = TileNameT % ( "N", lat, "E" if lon >= 0 else "W", abs(lon) )
			print "Current tile: %s" % ( TileName )
			if os.access( SRTMPath + TileName + SRTMExt, os.F_OK ) or os.access( SRTMPath + TileName + SRTMExt + ZipExt, os.F_OK ):
				for y in range( SRTMRows ):
					for x in range( SRTMCols ):
						# Write to the master points file and the next stage's input file
						if Alternate:
							if Geoid:
								Inp.write( "%.7f %.7f\n" % (lat + float(y + 0.5)/(SRTMCols - 1), lon + float(x + 0.5)/(SRTMRows - 1)) )
								Outp.write( "%s,%04d,%04d,%.7f,%.7f\n" % (TileName, y, x, lat + float(y + 0.5)/(SRTMCols - 1), lon + float(x + 0.5)/(SRTMRows - 1)) )
							else:
								Outp.write( "%s,%04d,%04d,%.7f,%.7f,%d\n" % (TileName, y, x, lat + float(y + 0.5)/(SRTMCols - 1), lon + float(x + 0.5)/(SRTMRows - 1), InvHeight) )
						else:
							if Geoid:
								Inp.write( "%.7f %.7f\n" % (lat + float(y)/(SRTMCols - 1), lon + float(x)/(SRTMRows - 1)) )
								Outp.write( "%s,%04d,%04d,%.7f,%.7f\n" % (TileName, y, x, lat + float(y)/(SRTMCols - 1), lon + float(x)/(SRTMRows - 1)) )
							else:
								Outp.write( "%s,%04d,%04d,%.7f,%.7f,%d\n" % (TileName, y, x, lat + float(y)/(SRTMCols - 1), lon + float(x)/(SRTMRows - 1), InvHeight) )

		# Close the master points file and the next stage's input file
		if Geoid:
			Inp.close()
		Outp.close()

		if Geoid:
			# Create the geoid heights file
			print "Calling GeoidEval to evaluate the Geoid heights ..."
			subprocess.call( [ "GeoidEval", "--input-file", DataIName, "--output-file", DataOName ] )

			os.remove( DataIName )
			os.rename( DataOName, DataIName )

			# Copy the geoid heights into the master points file
			print "Copying the Geoid heights into the points file ..."
			Inp = open( DataIName, "r" )
			Outp = open( DataOName, "w" )

			for line in open( DataFName, "r" ):
				# print	"Line:%s" % ( line )
				line	= line.strip().split(",")
				gH		= float( Inp.readline().strip() )
				Outp.write( ("%s,%s,%s,%s,%s,%." + format("%d" % Accuracy) + "f\n") % (line[0], line[1], line[2], line[3], line[4], gH) )

			Outp.close()
			Inp.close()

			os.remove( DataIName )
			DataFName = os.path.splitext( os.path.basename(DataFName) )[0]
			if os.access(DataFName + ptsExt, os.F_OK) and os.access(DataFName + bakExt, os.F_OK):
				os.remove( DataFName + bakExt )
			if os.access( DataFName + ptsExt, os.F_OK ):
				os.rename( DataFName + ptsExt, DataFName + bakExt )
			os.rename( DataOName, DataFName + ptsExt )

main()
