/*~********************************************************************

File:	OS2SRTM2.js

Charles Macfarlane, May/June 2015

Second of five programs/scripts to convert OS Landform Panorama or Terrain 50 elevation data to SRTM
format - the scripts are:
	OS2SRTM1.py		Begins by making for each possible SRTM cell a WGS84 coordinate line in a *.pt? file
					Requires GeoidEval from GeographicLib to be on the system path: http://geographiclib.sourceforge.net/
Either:
	GridInQuest		External OS program translates the WSG84 coordinates from above into OSGB36 and appends them
					http://www.ordnancesurvey.co.uk/business-and-government/help-and-support/navigation-technology/os-net/grid-inquest.html
Or:
	OS2SRTM2.js		Translates the WSG84 coordinates from above into OSGB36 and appends them
	(this script)	Unlike GIQ, this script cannot convert geiod heights, and therefore the /G switch cannot be
					used in the final script
	OS2SRTM3.py		Looks up heights from Landform Panorama and Terrain 50 and appends them
					http://www.ordnancesurvey.co.uk/business-and-government/products/land-form-panorama.html
					http://www.ordnancesurvey.co.uk/business-and-government/products/terrain-50.html
	OS2SRTM4.py 	Optionally uses a bitmap OSBitMap.bin to remove points corresponding to sea
	OS2SRTM5.py		Creates SRTM tiles

This second script takes the output from the first in the following format ...
	<SRTM TileName>	For example N55W006
	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
	<Tile Col>		Ditto
	<Lat>			Latitude in decimal degrees corresponding to row above
	<Lon>			Longitude in decimal degrees corresponding to col above
	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
... for example for N59W003.pt1 ...
	N59W003,0000,0000,59.0000000,-3.0000000,52.465
... and appends the UK East,North conversions of the above Lon,Lat as follows ...
	<SRTM TileName>	For example N55W006
	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
	<Tile Col>		Ditto
	<Lat>			Latitude in decimal degrees corresponding to row above
	<Lon>			Longitude in decimal degrees corresponding to col above
	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
	<East>			Easting in metres - translated from above (Lat,Lon,Ht) by Grid In Quest
	<North>			Northing in metres - ditto
	<EGM96 v OD>	-32768 (invalid value)
	<OD Name>		"Invalid"
... for example for N59W003.pt1 ...
	N59W003,0000,0000,59.0000000,-3.0000000,52.465,342647,1012974,-32768,Invalid

Then run the remaining scripts in the numerical order listed above, resuming with OS2SRTM3.py

Command-line switches:
/A	Use alternative cell centering and correspondingly process *.pt2 file(s)
	(*.pt1 = normal, *.pt2 = alternative - also see note 1)
/S	Do the six tiles covering the Shetlands only
	(at latitude 59N, ignore corrections at or below SRTM3 row 540, SRTM1 row 1620 - also see note 2)
/T	Do a test run across the middle of the country
	(see note 2)
/1	Create SRTM1 tiles, default is SRTM3
	(for compatibility only, doesn't need to do anything here - see also note 1)

Notes:

1	The /A and /1 switches MUST be used consistently throughout all the scripts.
2	The /S and /T switches can be introduced at any stage, but MUST be used consistently thereafter. These
	switches are mutually exclusive, if both are given, whichever is given second is ignored.
3	The /G switch in the final script is valid only if Grid InQuest is used instead of this script at stage 2.
4	Debug output is often best redirected to file, because when output is voluminous the need to scroll it
	slows program execution.

***********************************************************************

Charles Macfarlane's Work is licensed under a Creative Commons Licence -
Attribution-Non-Commercial-No Derivative Works 2.0 UK: England & Wales
http://creativecommons.org/licenses/by-nc-nd/2.0/uk/

For further details, please see:
http://www.macfh.co.uk/JavaJive/JavaJive.html

***********************************************************************

This script expects to be run by Windows Scripting Host, CScript.exe
This may not be installed by default on a Windows machine.  To install
WSH (you will likely need administrator rights), see ...
http://www.microsoft.com/downloads/details.aspx?FamilyId=C717D943-7E4B-4622-86EB-95A22B832CAA

For other OSs or interpreters, search for comments which begin ...
Non WSH use:
... currently there are 6, and make appropriate alterations.

(For best layout Tabs expected every 4 columns)

***********************************************************************

Updates:

**********************************************************************/


//	Non WSH use: delete or change appropriately these 5 lines
//	Create normal WSH support objects for the wider purpose of the program
//	var	wshNetwork	= new ActiveXObject( "WScript.Network" );
	var	wshShell	= new ActiveXObject( "WScript.Shell" );
	var	wshFileSys	= new ActiveXObject( "Scripting.FileSystemObject" );
	var interpreter	= WScript.FullName;			//	The name of the script interpreter
	var thisScript	= WScript.ScriptName;		//	The name of the script

//	Logging variables used by enumObj
var	logInd	= 0;
var	logTab	= "\t\t\t\t\t\t\t\t\t\t";

//	Program data

	var Pr4FileName	= ".\\..\\..\\..\\Local\\Resources\\Scripts\\Proj4jsUK.js";
	var	BIFileName	= "OS2SRTM-BI-N%d";
	var	ShFileName	= "OS2SRTM-Shetland-N%d";
	var	TstFileName	= "OS2SRTM-Test";
	var	BIStartX	= -8;
	var	BIStartY	= 49;
	var	BIEndX		= 2;
	var	BIEndY		= 61;
	var	ShStartX	= -3;
	var	ShStartY	= 59;
	var	ShEndX		= 0;
	var	ShEndY		= 61;
	var	TstStartX	= -7;
	var	TstStartY	= 55;
	var	TstEndX		= 0;
	var	TstEndY		= 56;
	var PtsPath		= ".\\";
	var Pt1Ext		= ".pt1";
	var Pt2Ext		= ".pt2";
	var PtsExt		= Pt1Ext;
	var BakExt		= ".bak";
	var LogExt		= ".log";
	var TmpExt		= ".tmp";
	var	InvHeight	= -32768;
	var	InvText		= "Invalid";

//	File data
	var forRead		= 1;
	var forWrite	= 2;
	var forAppend	= 8;

//	Create the required command-line object heirarchy and control variables
	initObjects();
	var	oneArcSec	= false;
	var	alternate	= false;
	var	testing		= false;
	var	shetland	= false;
	var	debug		= false;
	var overwrite	= false;

//	Create command-line parameter objects
	var	onePar		= new KeyPar	( "1", "-1\tProduce 1 arc second output (default 3)" );
	var	altPar		= new KeyPar	( "A", "-A\tAlternative cell centering" );
	var	shetPar		= new KeyPar	( "S", "-S\tDo a run for the Shetlands only" );
	var	tstPar		= new KeyPar	( "T", "-T\tDo a test run" );
//	var	ovrPar		= new KeyPar	( "O", "-O\tAutomatically overwrite existing destination points files" );
	var	dbgPar		= new KeyPar	( "D", "-D\tTurn on some debugging messages" );
	var	hlpPar		= new KeyPar	( "{0,2}(\\?|H|Help)?", "?\tShow Help", null, null, null, KeyPar.prototype.tags + "?" );

//	Create the command-line object
	var	cmdLine		= new CmdLine(
									[ onePar, altPar, shetPar, tstPar, /* ovrPar, */ dbgPar, hlpPar ],
									true, null, null,
									"Creates files containing height adjustments from Ordnance Survey data for each SRTM point.\n"
								);
//	Non WSH use: change appropriately WScript.Arguments
//	Scan/parse the command-line
	if( cmdLine.scan(WScript.Arguments) && !hlpPar.getScand() )
		{
		var DataNameT	= BIFileName;
		var	startX		= BIStartX;
		var	startY		= BIStartY;
		var	endX		= BIEndX;
		var	endY		= BIEndY;

		var logFileName, logFile, Pr4File, Pr4;

		//	Compatibility only, doesn't need to do anything
		if( onePar.getScand() )
			oneArcSec	= true;

		if( altPar.getScand() )
			{
			alternate	= true;
			PtsExt		= Pt2Ext;
			}

		if( shetPar.getScand() && !testing )
			{
			shetland	= true;
			startX		= ShStartX
			startY		= ShStartY
			endX		= ShEndX
			endY		= ShEndY
			if( wshFileSys.FileExists(PtsPath + ShFileName.replace( "%d", "59") + PtsExt) )
				DataNameT	= ShFileName
			}

		if( tstPar.getScand() && !shetland )
			{
			testing		= true;
			startX		= TstStartX
			startY		= TstStartY
			endX		= TstEndX
			endY		= TstEndY
			if( wshFileSys.FileExists(PtsPath + TstFileName.replace( "%d", "59") + PtsExt) )
				DataNameT	= TstFileName
			}
/*
		if( ovrPar.getScand() )
			overwrite	= true;
*/
		if( dbgPar.getScand() )
			debug		= true;

		if( debug )
			{
			//	log( "Command-line scanned successfully!", logFile );
			//	Open the log file
			logFileName = WScript.scriptName.substr( 0, WScript.scriptName.lastIndexOf(".") ) + LogExt;
			logFile = wshFileSys.CreateTextFile( logFileName, true );
			}

		//	Open the Proj4js projections script and run it
		log( "Reading script file ...\n\t\"" + Pr4FileName + "\"", logFile );
		Pr4File	= wshFileSys.OpenTextFile( Pr4FileName, forRead );
		Pr4 = Pr4File.ReadAll();
		Pr4File.Close();
		log( "Running script file ...\n\t\"" + Pr4FileName + "\"", logFile );
		eval( Pr4 );
		EPSG4326	= new Proj4js.Proj( "EPSG:4326" );
		EPSG27700	= new Proj4js.Proj( "EPSG:27700" );

		//	Read and update points file for each line of latitude
		for( var lat = startY; lat < endY; lat++ )
			{
			var	DataFName	= PtsPath + DataNameT.replace( "%d", "" + lat + "");
			var tmpFName	= DataFName + TmpExt;
			var ptsFName	= DataFName + PtsExt;
			var tmpFile, ptsFile;

			log( "Updating points file " + ptsFName + " ...", logFile );

			//	Open the source points file
			ptsFile	= wshFileSys.OpenTextFile( ptsFName, forRead );
			//	Open the destination temporary points file
			if( wshFileSys.FileExists(tmpFName) )
				tmpFile = wshFileSys.OpenTextFile( tmpFName, forWrite );
			else
				tmpFile = wshFileSys.CreateTextFile( tmpFName, false );

			var count = 0
			while( !ptsFile.AtEndOfStream )
				{
				var line	= ptsFile.ReadLine().trim();
				if( debug )
					log( line, logFile );
				count++;
				if( line )
					{
					data = line.split(",")
					try
						{
						var	sx	= parseFloat( data[4] );
						var	sy	= parseFloat( data[3] );
//						if( contains(sx, sy) )
							{
							var ng	= Proj4js.transform( EPSG4326, EPSG27700, { x:sx, y:sy } );
							var	xe	= Math.round( ng.x );
							var	yn	= Math.round( ng.y );
							line	+= "," + xe.format(6) + "," + yn.format(7) + "," + InvHeight + "," + InvText;
							tmpFile.WriteLine( line );
							if( debug )
								log( line, logFile );
							}
						}
					catch(e)
						{
						if( debug )
							log( "Error " + e + " in file: " + ptsFName + ", line: " + count, logFile );
						}
					}
				}

			tmpFile.Close();
			ptsFile.Close();

			//	If necessary, delete any existing back up file
			var bakFName = DataFName + BakExt;
			if( wshFileSys.FileExists(bakFName) )
				{
				if( debug )
					log( "Deleting back up file ...\n\t\"" + bakFName + "\"", logFile );
				wshFileSys.DeleteFile( bakFName );
				}

			//	Rename the input points file to *.bak
//			if( debug )
				log( "Backing up points file ...\n\t\"" + ptsFName + "\"", logFile );
			ptsFile = wshFileSys.GetFile( ptsFName );
			ptsFile.Name = wshFileSys.GetFileName( bakFName );

			//	Rename the temp file to be the new points file
			if( debug )
				log( "Creating updated points file ...\n\t\"" + ptsFName + "\"", logFile );
			tmpFile = wshFileSys.GetFile( tmpFName );
			tmpFile.Name = wshFileSys.GetFileName( ptsFName );
			}

		//	Close the log file
		if( logFile )
			logFile.Close();

		}
	else
		//	Show a suitable error message and quit
		log	(
			cmdLine.getHelp()
				+ (cmdLine.getGood() != "" ? "\nCommand-line successfully scanned:\n" + cmdLine.getGood() : "")
				+ (cmdLine.getBad() != "" ? "\nCommand-line not scanned:\n" + cmdLine.getBad() : "")
			);

//	Non WSH use: change appropriately WScript.Quit
//	Quit
	WScript.Quit();


//	Convert Easting, Northing to 100km Grid Reference
function ENToGridRef( anEast, aNorth, nDig )
	{
	//	UK Ordnance Survey 100km grid square designations
	var	grid100	= 	[
						"SV", "SW", "SX", "SY", "SZ", "TV", "TW",
						"SQ", "SR", "SS", "ST", "SU", "TQ", "TR",
						"SL", "SM", "SN", "SO", "SP", "TL", "TM",
						"SF", "SG", "SH", "SJ", "SK", "TF", "TG",
						"SA", "SB", "SC", "SD", "SE", "TA", "TB",
						"NV", "NW", "NX", "NY", "NZ", "OV", "OW",
						"NQ", "NR", "NS", "NT", "NU", "OQ", "OR",
						"NL", "NM", "NN", "NO", "NP", "OL", "OM",
						"NF", "NG", "NH", "NJ", "NK", "OF", "OG",
						"NA", "NB", "NC", "ND", "NE", "OA", "OB",
						"HV", "HW", "HX", "HY", "HZ", "JV", "JW",
						"HQ", "HR", "HS", "HT", "HU", "JQ", "JR",
						"HL", "HM", "HN", "HO", "HP", "JL", "JM"
					];
	var	km100 = 100000, result = null;

	if( (nDig != undefined) && (nDig/2 == Math.round(nDig/2)) && (nDig >= 2) )
		nDig = (nDig - 2) / 2;
	else
		nDig = 3
	with( Math )
		result = grid100[ floor(anEast/km100) + 7*floor(aNorth/km100) ] + ( nDig ? floor( (anEast%km100) / pow( 10, 5 - nDig ) ).format(nDig) +  floor( (aNorth%km100) / pow( 10, 5 - nDig ) ).format(nDig) : "" );
	return result;
	}


//	Check whether within UK OS coverage
function isUK( lon, lat )
	{
	return isInside( -9.5, 3.63, 49.7, 61.6, lon, lat ) && !isIreland( lon, lat );
	}

//	Check whether within Ireland
function isIreland( lon, lat )
	{
	return isInside( -10.66, -5.42, 51.39, 55.27, lon, lat );
	}

//	Check whether within given limits
function isInside( wLimit, eLimit, sLimit, nLimit, aLon, aLat )
	{
	return (aLon >= wLimit) && (aLon <= eLimit) && (aLat >= sLimit) && (aLat <= nLimit);
	}

//	Check whether point inside polygon, default polygon defines UK exc NI
//	(adapted from OL v2.12)
function contains( x, y, polyPar )
	{
	function getXforY( y, x1, y1, x2, y2 )
		{
		return (y - y2) * ( (x2 - x1) / (y2 - y1) ) + x2;
		}

	var polygon, numSeg, start, end, crosses, x1, y1, x2, y2, cx, cy, result;
	if( polyPar )
		//	Parameter Polygon
		polygon = polyPar
	else
		//	UK + IoM Cookie Cutter Polygon
		polygon =	[
						[-7.0,49.6],
						[-5.1,49.7],
						[ 0.3,50.5],
						[ 1.1,50.7],
						[ 2.0,51.4],
						[ 2.2,52.6],
						[-0.3,61.1],
						[-1.3,61.1],
						[-6.6,59.3],
						[-9.0,58.0],
						[-9.1,57.3],
						[-6.7,55.4],
						[-6.2,55.4],
						[-5.1,54.5],
						[-5.1,53.7],
						[-5.2,52.4],
						[-6.0,51.8],
						[-7.0,49.6]
					];

	numSeg	= polygon.length - 1;
	crosses	= 0;
	for( var i = 0; i < numSeg; i++ )
		{
		start	= polygon[i];
		x1		= start[0];
		y1		= start[1];
		end		= polygon[i + 1];
		x2		= end[0];
		y2		= end[1];

		/**
		* The following conditions enforce five edge-crossing rules:
		*    1. points coincident with edges are considered result;
		*    2. an upward edge includes its starting endpoint, and
		*    excludes its final endpoint;
		*    3. a downward edge excludes its starting endpoint, and
		*    includes its final endpoint;
		*    4. horizontal edges are excluded; and
		*    5. the edge-ray intersection point must be strictly right
		*    of the point P.
		*/
		if( y1 == y2 )
			//	Horizontal edge
			{
			if( y == y1 )
				//	Point on horizontal line
				{
				if
					(
						x1 <= x2 && (x >= x1 && x <= x2)	//	Right or vert
						||
						x1 >= x2 && (x <= x1 && x >= x2)	//	Left or vert
					)
					//	Point on edge
					{
					crosses = -1;
					break;
					}
				}
			//	Ignore other horizontal edges
			continue;
			}
		cx = getXforY( y, x1, y1, x2, y2 );
		if( cx == x )
			{
			//	Point on line
			if	(
					y1 < y2 && (y >= y1 && y <= y2)	//	Upward
					||
					y1 > y2 && (y <= y1 && y >= y2)	//	Downward
				)
				//	Point on edge
				{
				crosses = -1;
				break;
				}
			}
		if( cx <= x )
			//	No crossing to the right
			{
			continue;
			}
		if( x1 != x2 && (cx < Math.min(x1, x2) || (cx > Math.max(x1, x2))) )
			//	No Crossing
			{
			continue;
			}
		if	(
				(y1 < y2) && (y >= y1 && y < y2)	//	Upward
				||
				(y1 > y2) && (y < y1 && y >= y2)	//	Downward
			)
			{
			++crosses;
			}
		}
	result = (crosses == -1) ?
					//	On edge
					1 :
					//	Even (out) or odd (in)
					!!(crosses & 1);

	return result;
	}


//	Non WSH use: change appropriately wshFileSys.GetFileName
//	Initialise the command-line object heirarchy
//
//	Note:	MUST be called before creating any *Par (parameter) or CmdLine (command-line) objects.
//
function initObjects()
	{
	//	Improved Object printout
	Object.prototype.toString = function()
		{
		var	result	= "";
		for( var i in this )
			result += (result ? ", " : "" ) + i + ":" + this[i];
		return "{ " + result + " }";
		}

	//	Improved Array printout
	Array.prototype.toString = function ()
		{
		var result	= "";
		for( var i = 0, I = this.length; i < I; i++ )
			result += (result ? ", " : "" ) + i + ":" + this[i];
		return "[ " + result + " ]";
		}

	//	Extend the String object to trim leading and trailing whitespace
	String.prototype.trim = function()
		{
		return this.replace(/(^\s+)|(\s+$)/g, "");
		}

	//	Extend the Math object to give some simple number formatting
	Number.prototype.format = function( sigs, decs )
		{
		var result = ( decs != undefined ? Math.round( this * Math.pow(10, decs) ) / Math.pow(10, decs) : this );
		var sign = ( this < 0 ? "-" : "" );
		result = Math.abs( result ).toString().trim();
		if( decs )
			{
			if( result.indexOf(".") == -1 )
				result += ".";
			while( result.substr(result.indexOf(".") + 1).length < decs)
				result += "0";
			}
		while( (result.indexOf(".") != -1 ? result.indexOf(".") : result.length) < sigs )
			result = "0" + result;
		return sign + result;
		}

	FileSpecPar.prototype	= new Param;
	KeyPar.prototype		= new Param;
	KeyPar.prototype.tags	= "-/";				//	Windows users may prefer "-/"
	KeyPar.prototype.cases	= false;
	KeyStrPar.prototype		= new KeyPar;
	KeyLstPar.prototype		= new KeyStrPar;
	KeyTogPar.prototype		= new KeyStrPar;
	KeyTogPar.prototype.vals = "-+01FT";
	KeyIntPar.prototype		= new KeyStrPar;
	KeyFltPar.prototype		= new KeyIntPar;
	CmdLine.prototype.tTitl	= thisScript;
	CmdLine.prototype.tSyn	= "Syntax:\n" + wshFileSys.GetFileName(interpreter) + " " + thisScript + " <parameter(s)>\n";
	CmdLine.prototype.tErr	= "Command-Line Error - Required parameter(s) missing or wrong syntax or value:\n";
	CmdLine.prototype.tUndl = "============================================================================";
	CmdLine.prototype.tPars = "Parameters:";
	}

//	A set of command-line parameters
//
//	Parameters:
//
//		aPars	An array of objects each derived from Param and holding one recognised command-line parameter
//		bHelp	Boolean for whether to give help when errors are encountered scanning script arguments
//		sHelp	Either: A full (console) page of help for the user constructed by the programmer ...
//		... 	Or:		Null, in which case console help will be constructed from the parameters' help and:
//		STitle	Program title
//		sDesc	Program description
//		sSyntax	A line describing the command-line syntax
//		sEg		An example command line
//		sPars	An optional header for the list of help-lines for the individual command-line parameters
//
function CmdLine( aPars, bHelp, sHelp, sTitle, sDesc, sSyntax, sEg, sPars )
	{
	this.pars		= aPars ? aPars : null;
	this.bHelpful	= bHelp ? bHelp : false;
	this.help		= sHelp ? sHelp : null;
	if(	sTitle )
		this.tTitl	= sTitle;
	this.desc		= sDesc ? sDesc : null;
	if(	sSyntax )
		this.tSyn	= sSyntax;
	this.eg			= sEg;
	if(	sPars )
		this.tPars	= sPars;
	this.goodScans	= "";
	this.badScans	= "";
	this.getGood	= gGood;
	this.getBad		= gBad;
	this.scan		= scanCmdLine;
	this.getHelp	= fHelp;
	}

function scanCmdLine( args )
	{
	var	result	= true;
	var	subResult;
	var	msg		= this.tErr;
	for( var a = 0, A = args.length; a < A; a++ )
		{
//	Non WSH use: change appropriately to args[a]
		thisArg	= args(a);
		subResult = false;
		for( var p = 0, P = this.pars.length; p < P; p++ )
			{
			if( debug )
				log( "Parameter " + p + ": " + (this.pars[p].getHelp() ? this.pars[p].getHelp() : "") );
			if( this.pars[p].scan( thisArg ) )
				{
				subResult = true;
				this.goodScans += (this.goodScans != "" ? " " : "") + thisArg;
				if( debug )
					log( "Argument $" + a  + " " + thisArg + " scanned as " + this.pars[p].getValue() );
				break;
				}
			}
		if( (!subResult) && (this.bHelpful || debug) )
			{
			this.badScans += (this.badScans != "" ? " " : "") + thisArg;
			if( debug )
				log( "Argument $" + (a+1)  + " " + thisArg + " not scanned" );
			}
		result &= subResult;
		}
	subResult = true;
	for( p = 0, P = this.pars.length; p < P; p++ )
		if( this.pars[p].getReqd() && !this.pars[p].getScand() )
			{
			msg += this.pars[p].getHelp() + "\n";
			subResult = false;
			}
	if( (!subResult) && (this.bHelpful || debug) )
		log( msg + "\n" );
	result &= subResult;
	return result;
	}

function fHelp()
	{
	var	result = null;
	if( this.help != null )
		result = this.help;
	else
		{
		result = ( this.tTitl ? this.tTitl + "\n" + this.tUndl.substr(0, this.tTitl.length) + "\n\n" : "" )
					+ ( this.desc ? this.desc + "\n" : "" )
					+ ( this.tSyn ? this.tSyn + "\n" : "" )
					+ ( this.eg ? this.eg + "\n" : "" )
					+ ( this.tPars && this.pars && this.pars.length ? this.tPars + "\n" : "" );
		for( var p = 0, P = this.pars.length; p < P; p++ )
			if( this.pars[p].getHelp() )
				result += this.pars[p].getHelp() + "\n";
		}
	return result;
	}

function gGood()
	{
	return this.goodScans;
	}

function gBad()
	{
	return this.badScans;
	}

//	Base object for all command-line parameters
//
//	Parameters:
//
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		oValue	The initial value, defaults to null
//
function Param( sHelp, bReqd, oValue )
	{
	this.help		= sHelp ? sHelp : null;
	this.reqd		= bReqd ? true : false;
	this.scand		= false;
	this.value		= oValue ? oValue : null;
	this.getHelp	= gHelp;
	this.getReqd	= gReqd;
	this.getScand	= gScand;
	this.getValue	= gVal;
	this.setValue	= sVal;
	this.addListen	= aList;
	this.delListen	= dList;
	this.listeners	= new Array(0);
	}

function gHelp()
	{
	return this.help;
	}

function gReqd()
	{
	return this.reqd;
	}

function gScand()
	{
	return this.scand;
	}

function gVal()
	{
	return this.value;
	}

function sVal( nValue )
	{
	var	oValue = this.value;
	this.value = nValue;
	if( nValue != oValue )
		for( var l = 0, L = this.listeners.length; l < L; l++ )
			this.listeners[l]( nValue, oValue );
	return oValue;
	}

function aList( fListener )
	{
	this.listeners.push( fListener );
	}

function dList( fListener )
	{
	for( var l = this.listeners.length - 1; l >= 0; l-- )
		if( this.listeners[l] == fListener )
			this.listeners.splice( l , 1 );
	}

//	A file spec command-line parameter object
//
//	Format:	<filespec>
//
//	Note:	Untagged/unkeyed string parameters such as FileSpecPars must follow all other parameter types
//			in the aPars array passed to the CmdLine constructor.
//
//	Parameters:
//
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		sValue	The initial value, defaults to null
//		bStrict	Whether the parser should perform strict scanning, default is false
//				(would accept legal filenames that look like other parameters, such as -A)
//
function FileSpecPar( sHelp, bReqd, sValue, bStrict )
	{
	//	Object inheritance
	this.base	= Param;
	this.base( sHelp, bReqd, sValue );

	this.strict	= bStrict ? true : false;
	this.scan	= scanFileSpec;
	}

function scanFileSpec( aValue )
	{
	var	result 	= false;
	var	re 		= new RegExp( "[" + KeyPar.prototype.tags + "]" );
	if( !this.scand )
		{
		if( this.strict || !re.test(aValue.substr(0, 1)) )
			{
			this.setValue( aValue );
			this.scand	= true;
			result = true;
			}
		}
	return result;
	}

//	A keyed command-line parameter to toggle (switch or logically NOT) a boolean value
//
//	Format: [-]<key>
//
//	Parameters:
//
//		sKey	The string, usually a single character but can be a regular expression, to use as key
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		bValue	The initial value, defaults to false
//		bCase	Boolean for whether the key is case-sensitive, default is false
//		sTag	String of allowable tag characters, default is "-"
//
function KeyPar( sKey, sHelp, bReqd, bValue, bCase, sTag )
	{
	//	Object inheritance
	this.base	= Param;
	this.base( sHelp, bReqd, bValue ? bValue : false );

	this.key	= sKey;
	if( bCase )
		this.cases = true;
	if( sTag )
		this.tags	= sTag;
	this.scan	= scanKey;
	this.re		= new RegExp( "^[" + this.tags + "]" + this.key + "$", this.cases ? "g" : "ig" );
	}

function scanKey( aValue )
	{
	var	result = false;
	if( (!this.scand) && this.re.test(aValue) )
		{
		this.setValue( !this.getValue() );
		this.scand	= true;
		result = true;
		}
	return result;
	}

//	A keyed command-line string parameter
//
//	Format: [-]<key>:<string>
//
//	Parameters:
//
//		sKey	The string, usually a single character but can be a regular expression, to use as key
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		sValue	The initial value, defaults to null
//		bCase	Boolean for whether the key is case-sensitive, default is false
//		sTag	String of allowable tag characters, default is "-"
//
function KeyStrPar( sKey, sHelp, bReqd, sValue, bCase, sTag )
	{
	//	Object inheritance
	this.base	= KeyPar;
	this.base( sKey, sHelp, bReqd, sValue ? sValue : null, bCase, sTag );

	this.scan	= scanKeyStr;
	this.re		= new RegExp( "^[" + this.tags + "]" + this.key + ":?(.+)$", this.cases ? "g" : "ig" );
	}

function scanKeyStr( aValue )
	{
	var	result = false;
	if( !this.scand )
		{
		try
			{
			this.setValue( this.re.exec(aValue)[1] );
			this.scand	= true;
			result = true;
			}
		catch(e)
			{
			if( debug )
				log( "scanKeyStr - exception caught: 0x" + ((e.number+0xFFFFFFFF)%0xFFFFFFFF).toString(16).toUpperCase() + "\n\t" + e.description );
			}
		}
	return result;
	}

//	A keyed command-line parameter to set a boolean value
//
//	Format: [-]<key>:<-+01>
//
//	Parameters:
//
//		sKey	The string, usually a single character but can be a regular expression, to use as key
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		bValue	The initial value, defaults to false
//		bCase	Boolean for whether the key and the toggle characters are case-sensitive, default is false
//		sTag	String of allowable tag characters, default is "-"
//		sVals	String of allowable toggle characters, should be in pairs "<False><True>", as in default "-+01FT"
//
function KeyTogPar( sKey, sHelp, bReqd, bValue, bCase, sTag, sVals )
	{
	//	Object inheritance
	this.base	= KeyStrPar;
	this.base( sKey, sHelp, bReqd, bValue ? bValue : false, bCase, sTag );

	if( sVals )
		this.vals = sVals;
	this.scan	= scanKeyTog;
	this.re		= new RegExp( "^[" + this.tags + "]" + this.key + ":?([" + this.vals + "])?$", this.cases ? "g" : "ig" );
	}

function scanKeyTog( aValue )
	{
	var	result = false;
	if( !this.scand )
		{
		try
			{
			var	tog = this.re.exec(aValue);
			if( tog != null )
				{
				tog = tog[tog.length - 1];
				switch( tog ? this.vals.indexOf( this.cases ? tog : tog.toUpperCase() ) % 2 : -1 )
					{
					case -1:	this.setValue( !this.getValue() );
								this.scand	= true;
								result = true;
								break;

					case 0:		this.setValue( false );
								this.scand	= true;
								result = true;
								break;

					case 1:		this.setValue( true );
								this.scand	= true;
								result = true;
								break;

					default:	break;
					}
				}
			}
		catch(e)
			{
			if( debug )
				log( "scanKeyTog - exception caught: 0x" + ((e.number+0xFFFFFFFF)%0xFFFFFFFF).toString(16).toUpperCase() + "\n\t" + e.description );
			}
		}
	return result;
	}

//	A keyed command-line parameter to choose from a list of allowed strings
//
//	Format: [-]<key>:<string>
//
//	Parameters:
//
//		sKey	The string, usually a single character but can be a regular expression, to use as key
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		sVals	Array of allowed string choices
//		bReqd	Boolean for whether the parameter must be present, default is false
//		bValue	The initial value, defaults to false
//		bCase	Boolean for whether the key and the string values are case-sensitive, default is false
//		sTag	String of allowable tag characters, default is "-"
//
function KeyLstPar( sKey, sHelp, sVals, bReqd, bValue, bCase, sTag )
	{
	//	Object inheritance
	this.base	= KeyStrPar;
	this.base( sKey, sHelp, bReqd, bValue ? bValue : false, bCase, sTag );

	if( sVals )
		this.vals = sVals;
	this.scan	= scanKeyLst;
	this.re		= new RegExp( "^[" + this.tags + "]" + this.key + ":?(" + this.vals.join("|") + ")$", this.cases ? "g" : "ig" );
	}

function scanKeyLst( aValue )
	{
	var	result = false;
	if( !this.scand )
		{
		try
			{
			this.setValue( this.re.exec(aValue)[1] );
			this.scand	= true;
			result = true;
			}
		catch(e)
			{
			if( debug )
				log( "scanKeyStr - exception caught: 0x" + ((e.number+0xFFFFFFFF)%0xFFFFFFFF).toString(16).toUpperCase() + "\n\t" + e.description );
			}
		}
	return result;
	}

//	A keyed command-line bounded integer parameter
//
//	Format: [-]<key>:<decimal>|B<binary>|O<octal>|H<hexadecimal>
//
//	Parameters:
//
//		sKey	The string, usually a single character but can be a regular expression, to use as key
//		iLower	The lower bound, defaults to javascript's smallest integer -2^52
//		iUpper	The upper bound, defaults to javascript's largest integer +2^52
//		iValue	The initial value, defaults to null
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		bCase	Boolean for whether the key, B, O, H are case-sensitive, default is false
//		sTag	String of allowable tag characters, default is "-"
//
function KeyIntPar( sKey, iLower, iUpper, iValue, sHelp, bReqd, bCase, sTag )
	{
	//	Object inheritance
	this.base	= KeyStrPar;
	this.base( sKey, sHelp, bReqd, iValue, bCase, sTag );

	this.lower		= iLower != null ? iLower : -Math.pow(2,52);
	this.upper		= iUpper != null ? iUpper : Math.pow(2,52);
	this.getLower	= gLower;
	this.setLower	= sLower;
	this.getUpper	= gUpper;
	this.setUpper	= sUpper;
	this.scan		= scanKeyInt;
	this.re			= new RegExp( "^[" + this.tags + "]" + this.key + ":?(([-+]?[0-9]+)|(B[-+]?[0-1]+)|(O[-+]?[0-7]+)|(H[-+]?[0-9A-F]+))$", this.cases ? "g" : "ig" );
	}

function scanKeyInt( aValue )
	{
	var	result = false;
	if( !this.scand )
		{
		try
			{
			var	base;
			var	numb = this.re.exec(aValue)[1];
			switch( numb.charAt(0).toUpperCase() )
				{
				default:	base = 10;
							break;

				case "B":	numb = numb.substr( 1 );
							base = 2;
							break;

				case "O":	numb = numb.substr( 1 );
							base = 8;
							break;

				case "H":	numb = numb.substr( 1 );
							base = 16;
							break;
				}
			numb = parseInt( numb, base );
			if( (numb >= this.lower) && (numb <= this.upper) )
				{
				this.setValue( numb );
				this.scand	= true;
				result = true;
				}
			}
		catch(e)
			{
			if( debug )
				log( "scanKeyInt - exception caught: 0x" + ((e.number+0xFFFFFFFF)%0xFFFFFFFF).toString(16).toUpperCase() + "\n\t" + e.description );
			}
		}
	return result;
	}

function gLower()
	{
	return this.lower;
	}

function sLower( nLower )
	{
	var	oLower = this.lower;
	this.lower = nLower;
	return oLower;
	}

function gUpper()
	{
	return this.upper;
	}

function sUpper( nUpper )
	{
	var	oUpper = this.upper;
	this.upper = nUpper;
	return oUpper;
	}

//	A keyed command-line bounded floating point parameter accepting a number in decimal or scientific format
//
//	Format: [-]<key>:<fp decimal>
//
//	Parameters:
//
//		sKey	The string, usually a single character but can be a regular expression, to use as key
//		iLower	The lower bound, defaults to javascript's smallest number -2^52
//		iUpper	The upper bound, defaults to javascript's largest number +2^52
//		iValue	The initial value, defaults to null
//		sHelp	Help text (preferably < 79 characters) to prompt user
//		bReqd	Boolean for whether the parameter must be present, default is false
//		bCase	Boolean for whether the key and the E denoting an exponent are case-sensitive, default is false
//		sTag	String of allowable tag characters, default is "-"
//
function KeyFltPar( sKey, iLower, iUpper, iValue, sHelp, bReqd, bCase, sTag )
	{
	//	Object inheritance
	this.base	= KeyIntPar;
	this.base( sKey, iLower, iUpper, iValue, sHelp, bReqd, bCase, sTag );

	this.scan		= scanKeyFlt;
	this.re			= new RegExp( "^[" + this.tags + "]" + this.key + ":?([-+]?[.0-9]+)(?:E([-+]?[.0-9]+))?$", this.cases ? "g" : "ig" );
	}

function scanKeyFlt( aValue )
	{
	var	result = false;
	if( !this.scand )
		{
		try
			{
			var	comps = this.re.exec(aValue);
			var	numb = comps[2] ? parseFloat( comps[1] )*Math.pow( 10, parseFloat(comps[2]) ) : parseFloat( comps[1] );
			if( (numb >= this.lower) && (numb <= this.upper) )
				{
				this.setValue( numb );
				this.scand	= true;
				result = true;
				}
			}
		catch(e)
			{
			if( debug )
				log( "scanKeyFlt - exception caught: 0x" + ((e.number+0xFFFFFFFF)%0xFFFFFFFF).toString(16).toUpperCase() + "\n\t" + e.description );
			}
		}
	return result;
	}


//	Non WSH use: change appropriately WScript.Echo, TextStream.WriteLine
//	Output logging messages
function log( _msg, _logfile )
	{
	WScript.Echo( _msg );
	if( _logfile != null )
		_logfile.WriteLine( _msg );
	return;
	}
