#!python
#
# OS2SRTM3.py
#
# Charles Macfarlane, May/June 2015
#
# Third of five programs/scripts to convert OS Landform Panorama or Terrain 50 elevation data to SRTM format -
# the programs/scripts are:
#	OS2SRTM1.py		Begins by making for each possible SRTM cell a WGS84 coordinate line in a *.pt? file
#					Requires GeoidEval from GeographicLib to be on the system path: http://geographiclib.sourceforge.net/
# Either:
#	GridInQuest		External OS program translates the WSG84 coordinates from above into OSGB36 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/help-and-support/navigation-technology/os-net/grid-inquest.html
# Or:
#	OS2SRTM2.js		Translates the WSG84 coordinates from above into OSGB36 and appends them
#					Unlike GIQ, this script cannot convert geiod heights, and therefore the /G switch cannot be
#					used in the final script
#	OS2SRTM3.py		Looks up heights from Landform Panorama and Terrain 50 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/products/land-form-panorama.html
#					http://www.ordnancesurvey.co.uk/business-and-government/products/terrain-50.html
#	OS2SRTM4.py		Optionally uses a bitmap OSBitMap.bin to remove points corresponding to sea
#	OS2SRTM5.py		Creates SRTM tiles
#
# This third script takes the GIQ transformed *.pt1 or *.pt2 CSV output of the previous one and appends
# Landform Panorama and Terrain 50 height data to each point line, resulting in ...
#	<SRTM TileName>	For example N55W006
#	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
#	<Tile Col>		Ditto
#	<Lat>			Latitude in decimal degrees corresponding to row above
#	<Lon>			Longitude in decimal degrees corresponding to col above
#	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
#	<East>			Easting in rounded metres - translated from above (Lat,Lon,Ht) by Grid In Quest
#	<North>			Northing in rounded metres - ditto
#	<EGM96 v OD>	Height in metres of the EGM96 geoid above Ordnance Datum - ditto
#	<OD Name>		Name of above Ordnance Datum
#	<LP Height>		Terrain height in metres given by OS Landform Panorama (uncorrected v EGM96)
#	<T50 Height>	Terrain height in metres given by OS Terrain 50 (uncorrected v EGM96)
# ... for example for N59W003.pt1 ...
#	N59W003,0000,0000,59.0000000,-3.0000000,52.465,342647,1012974,0.874,Newlyn (Orkney Isles only),28.183,28.441
#
# Command-line switches:
#	/D	Produce debug output
#	/A	Use alternative cell centering and correspondingly process *.pt2 file(s)
#		(*.pt1 = normal, *.pt2 = alternative - also see note 1)
#	/S	Do the six tiles covering the Shetlands only
#		(at latitude 59N, ignore corrections at or below SRTM3 row 540, SRTM1 row 1620 - also see note 2)
#	/T	Do a test run across the middle of the country
#		(see note 2)
#	/1	Create SRTM1 tiles, default is SRTM3
#		(see note 1, though here it's actually ignored as it makes no difference)
#
# Notes:
#
#	1	The /A and /1 switches MUST be used consistently throughout all the scripts.
#	2	The /S and /T switches can be introduced at any stage, but MUST be used consistently thereafter. These
#		switches are mutually exclusive, if both are given, whichever is given second is ignored.
#	3	The /G switch in the final script is only valid if Grid InQuest was used at stage 2.
#	4	Debug output is often best redirected to file, because when output is voluminous the need to scroll it
#		slows program execution.
#
# ***********************************************************************
#
# Charles Macfarlane's Work is licensed under a Creative Commons Licence -
# Attribution-Non-Commercial-No Derivative Works 2.0 UK: England & Wales
# http://creativecommons.org/licenses/by-nc-nd/2.0/uk/
#
# For further details, please see:
# http://www.macfh.co.uk/JavaJive/JavaJive.html
#
# ***********************************************************************


import os
import array
import re
import sys
import zipfile
from string import lower, upper
from math import ceil,floor,radians
from datetime import datetime

PtsPath		= ".\\"
SRTMPath	= os.path.join( "..", "" )
OSLPPath	= os.path.join( "..", "..", "OS", "Landform Panorama", "DTM", "ASCII", "data", "" )
OST50Path	= os.path.join( "..", "..", "OS", "Terrain 50", "data", "" )
Pts1Ext		= ".pt1"
Pts2Ext		= ".pt2"
bakExt		= ".bak"
BIFileName	= "OS2SRTM-BI-N%d"
ShFileName	= "OS2SRTM-Shetland-N%d"
TstFileName	= "OS2SRTM-Test"
BIStartX	= -8
BIStartY	= 49
BIEndX		= 2
BIEndY		= 61
ShStartX	= -3
ShStartY	= 59
ShEndX		= 0
ShEndY		= 61
TstStartX	= -7
TstStartY	= 55
TstEndX		= 0
TstEndY		= 56
startX		= BIStartX
startY		= BIStartY
endX		= BIEndX
endY		= BIEndY

# Number of decimal places for the heights
Accuracy = 3

# Debug flag
Debug		= False
# Flags which cell alignment method to use:
#	False	Use SW corner of SRTM and OS Landform Panorama cells, adjust to middle of Terrain 50 cells
#	True	Adjust to middle of SRTM and OS Landform Panorama cells, use SW corner of Terrain 50 cells
Alternate	= False
# Shetland flag
Shetland	= False
# Test flag
Test		= False


# Ordnance Survey Landform Panorama (OSLP) elevation tile:
#	http://www.ordnancesurvey.co.uk/business-and-government/products/land-form-panorama.html
# Format: 401 x 401 pixels per 20km2 tile, X east within Y north, starting from north west corner.
class OSLPTile:

	# Tile width (sample points)
	ncols	= 401
	# Tile height (sample points)
	nrows	= 401
	# Lower left cell easting (metres)
	xll		= 0
	# Lower left cell northing (metres)
	yll		= 0
	# Cell size in metres
	cell	= 50
	# Terrain Height Data
	data	= None
	# Time of last access
	access	= 0

	def __init__( self, cols, rows, x, y, cell, data ):
		if isinstance( cols, int ):
			self.ncols	= cols
			self.nrows	= rows
			self.xll	= x
			self.yll	= y
			self.cellsize	= cell
			self.data	= data
		else:
			start		= GridRefToEN( cols )
			self.xll	= start.x
			self.yll	= start.y
			self.data	= None
		self.access	= datetime.now()

	# Tile width
	def getCols( self ):
		return self.ncols

	# Tile height
	def getRows( self ):
		return self.nrows

	# Lower left cell easting
	def getXLL( self ):
		return self.xll

	# Lower left cell northing
	def getYLL( self ):
		return self.yll

	# Cell size in metres
	def getCell( self ):
		return self.cell

	# Terrain Height Data
	def getData( self ):
		return self.data

	# Terrain Height Data
	def getAccess( self ):
		return self.access

	# Log access time
	def setAccess( self, sTime ):
		if sTime:
			self.access	= sTime
		else:
			self.access	= datetime.now()


class OSLPTileSet:

	# Terrain height file extensions
	ext		= ".asc"
	# Minimum height
	min		= -32767
	# Maximum height
	max		= 32767
	# Cell size in metres
	cell	= 50
	# Tile cache size
	cLim	= 13
	# OSLP tile cache
	tiles	= dict()

	def __init__( self ):
		if self.tiles == None:
			self.tiles	= dict()

	# Get cell size in metres
	def getCell( self ):
		return self.cell

	# Get tile data
	def getData( self, tile ):
		return self.tiles[tile].getData()

	# Calc OSLP filename
	def getOSLPName( self, east, north ):
		tn	= ENToGridRef( east, north )
		return "%s%1d%1d" % ( tn[0:2], 2*floor(int(tn[2:3])/2), 2*floor(int(tn[3:4])/2) )

	# Read OSLP file
	def getOSLPData( self, east, north ):
		global Debug
		tn = lower( self.getOSLPName(east, north) )
		if tn not in self.tiles:
			data = asc = None
			fn = os.path.join( OSLPPath, tn[0:2], tn + self.ext )
			if Debug:
				print "Looking for file %s ..." % fn
			try:
				asc = open( fn, "r" )
			except IOError:
				if Debug:
					print "WARNING:  Could not open Landform Panorama data file for tile: %s" % upper(tn)
			if asc != None:
				# print "/* Opened file %s */<br>\n" % fn
				for y in range( 5 ):
					try:
						li = asc.readline().strip().split()
						if re.compile( "^ncols$", re.IGNORECASE ).match( li[0] ):
							ncols	= int( li[1] )
						elif re.compile( "^nrows$", re.IGNORECASE ).match( li[0] ):
							nrows	= int( li[1] )
						elif re.compile( "^xllcenter$", re.IGNORECASE ).match( li[0] ):
							xll	= int( li[1] )
						elif re.compile( "^yllcenter$", re.IGNORECASE ).match( li[0] ):
							yll	= int( li[1] )
						elif re.compile( "^cellsize$", re.IGNORECASE ).match( li[0] ):
							cell	= int( li[1] )
						# print "/* li[0]: %s, li[1]: %s */<br>/n" % ( li[0], li[1] )
					except Exception:
						# print "/* Could not read or bad data %s in header line %d of file %s */<br>\n" %( li[1], y, fn )
						pass

				data	= [0.0] * ncols * nrows
				for y in range( nrows ):
					try:
						li	= asc.readline().strip().split()
						for x in range( ncols ):
							try:
								data[(nrows - 1 - y)*ncols + x] = float( li[x] )
							except Exception:
								# print "/* Bad data %s in line %d column %d of file %s */<br>\n" %( li[x], y + 5, x, fn )
								pass
					except Exception:
						# print "/* Could not read line %d of data file: %s */<br>\n" % (y, fn)
						pass
				asc.close()
			if data:
				# print "/* Loaded Tile: %s */<br>\n" % tn
				self.tiles[tn] = OSLPTile( ncols, nrows, xll, yll, cell, data )
			else:
				# print "/* Blank Tile: %s */<br>\n" % tn
				self.tiles[tn] = OSLPTile( tn, None, None, None, None, None )
			while len( self.tiles ) > self.cLim:
				oldest	= ""
				keys	= self.tiles.iterkeys()
				for t in keys:
					# print "/* Oldest: %s, This: %s */<br>\n" % ( oldest, t )
					if (oldest == "") or (self.tiles[t].getAccess() < self.tiles[oldest].getAccess()):
						oldest = t
				if oldest <> "":
					# print "/* Removing %s */<br>\n" % ( oldest )
					del( self.tiles[oldest] )
			# keys = self.tiles.iterkeys()
			# print "/*\nTile cache contents now (%d tiles):" % len( self.tiles )
			# for t in keys:
			#	print " %s" % ( t )
			# print "*/<br>\n"
		return self.tiles[tn]

	# Look up raw height
	def getCleanHt( self, x, y ):
		tile	= self.getOSLPData( x, y )
		ht	= 0.0
		if tile:
			tile.setAccess( None )
			if tile.getData():
				x	= int( round((x - tile.getXLL()) / tile.getCell()) )
				y	= int( round((y - tile.getYLL()) / tile.getCell()) )
				of	= y*tile.getCols() + x
				ht	= tile.getData()[of]
				# print "/* x:%u, y:%u, of:%u, ht:%d, len:%u */<br>\n" % ( x, y, of, ht, len(tile.getData()) )
		return ht

	# Look up and interpolate height
	def getHeight( self, x, y ):
		ht = 0.0
		cx = x / self.cell
		cy = y / self.cell
		x0 = floor( cx )
		y0 = floor( cy )
		x1 = x0 + 1
		y1 = y0 + 1
		dx = cx - x0;
		dy = cy - y0;

		# Possibly adjust for tile cell coordinates being of the middle, not the SW corner
		if Alternate:
			if dx < 0.5:
				if x0 > 0:	# Doubt if this will ever fail in practice
					x0 -= 1
				x1 -= 1
				dx += 0.5
			else:
				dx -= 0.5
			if dy < 0.5:
				if y0 > 0:	# Doubt if this will ever fail in practice
					y0 -= 1
				y1 -= 1
				dy += 0.5
			else:
				dy -= 0.5

		x0 *= self.cell
		y0 *= self.cell
		x1 *= self.cell
		y1 *= self.cell
		# print "/* x:%.0f, y:%.0f, cx:%.3f, cy:%.3f, x0:%u, y0:%u, x1:%u, y1:%u, dx:%.3f, dy:%.3f */<br>\n" % (x,y,cx,cy,x0,y0,x1,y1,dx,dy)
		hSW = self.getCleanHt( x0, y0 )
		hSE = self.getCleanHt( x1, y0 )
		hNW = self.getCleanHt( x0, y1 )
		hNE = self.getCleanHt( x1, y1 )
		hS = ( 1 - dx )*hSW + dx*hSE
		hN = ( 1 - dx )*hNW + dx*hNE
		ht = ( 1 - dy )*hS + dy*hN
		# print "/* x:%.0f, y:%.0f, cx:%.3f, cy:%.3f, x0:%u, y0:%u, x1:%u, y1:%u, dx:%.3f, dy:%.3f, hSW:%d, hSE:%d, hNW:%d, hNE:%d, hS:%.3f, hN:%.3f, ht:%.3f  */<br>\n" % (x,y,cx,cy,x0,y0,x1,y1,dx,dy,hSW,hSE,hNW,hNE,hS,hN,ht)
		return ht

# The current OSLP tiles
OSLP = OSLPTileSet()


# Ordnance Survey Terrain 50 (OST50) elevation tile:
#	http://www.ordnancesurvey.co.uk/business-and-government/products/terrain-50.html
# Format: 200 x 200 pixels per 10km2 tile, X east within Y north, starting from north west corner.
class OST50Tile:

	# Tile width (sample points)
	ncols	= 200
	# Tile height (sample points)
	nrows	= 200
	# Lower left cell easting (metres)
	xll		= 0
	# Lower left cell northing (metres)
	yll		= 0
	# Cell size in metres
	cell	= 50
	# Terrain Height Data
	data	= None
	# Time of last access
	access	= 0

	def __init__( self, cols, rows, x, y, cell, data ):
		if isinstance( cols, int ):
			self.ncols	= cols
			self.nrows	= rows
			self.xll	= x
			self.yll	= y
			self.cellsize	= cell
			self.data	= data
		else:
			start		= GridRefToEN( cols )
			self.xll	= start.x
			self.yll	= start.y
			self.data	= None
		self.access	= datetime.now()

	# Tile width
	def getCols( self ):
		return self.ncols

	# Tile height
	def getRows( self ):
		return self.nrows

	# Lower left cell easting
	def getXLL( self ):
		return self.xll

	# Lower left cell northing
	def getYLL( self ):
		return self.yll

	# Cell size in metres
	def getCell( self ):
		return self.cell

	# Terrain Height Data
	def getData( self ):
		return self.data

	# Terrain Height Data
	def getAccess( self ):
		return self.access

	# Log access time
	def setAccess( self, sTime ):
		if sTime:
			self.access	= sTime
		else:
			self.access	= datetime.now()


class OST50TileSet:

	# Terrain height file extensions
	ext		= ".asc"
	# Minimum height
	min		= -32767
	# Maximum height
	max		= 32767
	# Cell size in metres
	cell	= 50
	# Tile cache size
	cLim	= 13
	# OST50 tile cache
	tiles	= dict()
	# Zip file name template
	ZipFN	= "_OST50GRID_20130401.zip"

	def __init__( self ):
		if self.tiles == None:
			self.tiles	= dict()

	# Get cell size in metres
	def getCell( self ):
		return self.cell

	# Get tile data
	def getData( self, tile ):
		return self.tiles[tile].getData()

	# Calc OST50 filename
	def getOST50Name( self, east, north ):
		return ENToGridRef( east, north )

	# Read OST50 file
	def getOST50Data( self, east, north ):
		global Debug
		tn = self.getOST50Name( east, north )
		if tn not in self.tiles:
			data = asc = zip = None
			fn = os.path.join( OST50Path, tn[0:2].lower() )
			try:
				if Debug:
					print "Looking for file %s ..." % os.path.join(fn, tn.upper() + self.ext)
				asc = open( os.path.join(fn, tn.upper() + self.ext), "r" )
			except IOError:
				if Debug:
					print "Looking for file %s ..." % os.path.join(fn, tn.lower() + self.ZipFN)
				try:
					zip = zipfile.ZipFile( os.path.join(fn, tn.lower() + self.ZipFN), "r" )
					asc = zip.open( tn.upper() + self.ext )
				except IOError:
					if Debug:
						print "WARNING:  Could not open Terrain 50 data file for tile: %s" % tn
			if asc != None:
				# print "/* Opened file %s */<br>\n" % fn
				for y in range( 5 ):
					try:
						li = asc.readline().strip().split()
						if re.compile( "^ncols$", re.IGNORECASE ).match( li[0] ):
							ncols	= int( li[1] )
						elif re.compile( "^nrows$", re.IGNORECASE ).match( li[0] ):
							nrows	= int( li[1] )
						elif re.compile( "^xllcorner$", re.IGNORECASE ).match( li[0] ):
							xll	= int( li[1] )
						elif re.compile( "^yllcorner$", re.IGNORECASE ).match( li[0] ):
							yll	= int( li[1] )
						elif re.compile( "^cellsize$", re.IGNORECASE ).match( li[0] ):
							cell	= int( li[1] )
						# print "/* li[0]: %s, li[1]: %s */<br>/n" % ( li[0], li[1] )
					except Exception:
						# print "/* Could not read or bad data %s in header line %d of file %s */<br>\n" %( li[1], y, fn )
						pass

				data	= [0.0] * ncols * nrows
				for y in range( nrows ):
					try:
						li	= asc.readline().strip().split()
						for x in range( ncols ):
							try:
								data[(nrows - 1 - y)*ncols + x] = float( li[x] )
							except Exception:
								# print "/* Bad data %s in line %d column %d of file %s */<br>\n" %( li[x], y + 5, x, fn )
								pass
					except Exception:
						# print "/* Could not read line %d of data file: %s */<br>\n" % (y, fn)
						pass
				asc.close()
			if zip != None:
				zip.close()
			if data:
				# print "/* Loaded Tile: %s */<br>\n" % tn
				self.tiles[tn] = OST50Tile( ncols, nrows, xll, yll, cell, data )
			else:
				# print "/* Blank Tile: %s */<br>\n" % tn
				self.tiles[tn] = OST50Tile( tn, None, None, None, None, None )
			while len( self.tiles ) > self.cLim:
				oldest	= ""
				keys	= self.tiles.iterkeys()
				for t in keys:
					# print "/* Oldest: %s, This: %s */<br>\n" % ( oldest, t )
					if (oldest == "") or (self.tiles[t].getAccess() < self.tiles[oldest].getAccess()):
						oldest = t
				if oldest <> "":
					# print "/* Removing %s */<br>\n" % ( oldest )
					del( self.tiles[oldest] )
			# keys = self.tiles.iterkeys()
			# print "/*\nTile cache contents now (%d tiles):" % len( self.tiles )
			# for t in keys:
			#	print " %s" % ( t )
			# print "*/<br>\n"
		return self.tiles[tn]

	# Look up raw height
	def getCleanHt( self, x, y ):
		tile	= self.getOST50Data( x, y )
		ht	= 0.0
		if tile:
			tile.setAccess( None )
			if tile.getData():
				x	= int( round((x - tile.getXLL()) / tile.getCell()) )
				y	= int( round((y - tile.getYLL()) / tile.getCell()) )
				of	= y*tile.getCols() + x
				ht	= tile.getData()[of]
				# print "/* x:%u, y:%u, of:%u, ht:%d, len:%u */<br>\n" % ( x, y, of, ht, len(tile.getData()) )
		return ht

	# Look up and interpolate height
	def getHeight( self, x, y ):
		ht = 0.0
		cx = x / self.cell
		cy = y / self.cell
		x0 = floor( cx )
		y0 = floor( cy )
		x1 = x0 + 1
		y1 = y0 + 1
		dx = cx - x0;
		dy = cy - y0;

		# Possibly adjust for tile cell coordinates being of the SW corner, not the middle
		if not Alternate:
			if dx < 0.5:
				if x0 > 0:	# Doubt if this will ever fail in practice
					x0 -= 1
				x1 -= 1
				dx += 0.5
			else:
				dx -= 0.5
			if dy < 0.5:
				if y0 > 0:	# Doubt if this will ever fail in practice
					y0 -= 1
				y1 -= 1
				dy += 0.5
			else:
				dy -= 0.5

		x0 *= self.cell
		y0 *= self.cell
		x1 *= self.cell
		y1 *= self.cell
		# print "/* x:%.0f, y:%.0f, cx:%.3f, cy:%.3f, x0:%u, y0:%u, x1:%u, y1:%u, dx:%.3f, dy:%.3f */<br>\n" % (x,y,cx,cy,x0,y0,x1,y1,dx,dy)
		hSW = self.getCleanHt( x0, y0 )
		hSE = self.getCleanHt( x1, y0 )
		hNW = self.getCleanHt( x0, y1 )
		hNE = self.getCleanHt( x1, y1 )
		hS = ( 1 - dx )*hSW + dx*hSE
		hN = ( 1 - dx )*hNW + dx*hNE
		ht = ( 1 - dy )*hS + dy*hN
		# print "/* x:%.0f, y:%.0f, cx:%.3f, cy:%.3f, x0:%u, y0:%u, x1:%u, y1:%u, dx:%.3f, dy:%.3f, hSW:%d, hSE:%d, hNW:%d, hNE:%d, hS:%.3f, hN:%.3f, ht:%.3f  */<br>\n" % (x,y,cx,cy,x0,y0,x1,y1,dx,dy,hSW,hSE,hNW,hNE,hS,hN,ht)
		return ht

# The current OST50 tiles
OST50 = OST50TileSet()


# Cartographical geometry

# Class defining a 2D coordinate
class Duple:
	def __init__( self, xx, yy ):
		self.x = xx
		self.y = yy

# UK East,North to GridRef
def ENToGridRef( anEast, aNorth ):
	result	= ""
	pow10	= 100000
	anEast 	+= 10*pow10
	aNorth 	+= 5*pow10
	anEastH	= floor( anEast / pow10 )
	aNorthH	= floor( aNorth / pow10 )
	anEastL	= floor( anEast % pow10 )
	aNorthL	= floor( aNorth % pow10 )
	nDig	= 2
	for i in range( nDig ):
		eSq = nSq = ""
		if i == 0:
			eSq	= floor( anEastH / 5 )
			nSq	= floor( aNorthH / 5 )
		else:
			eSq	= anEastH % 5
			nSq	= aNorthH % 5
		square	= int( (5*( 4 - nSq ) + eSq) % 25 )
		if square > 7:
			square += 1
		result += chr( 65 + square )
	nDig	= 1
	pow10	= 10**(5 - nDig)
	if nDig > 0:
		nDig = "%d" % nDig
		nDig = "%0" + nDig + "d%0" + nDig + "d"
		result += nDig % ( floor(anEastL/pow10), floor(aNorthL/pow10) )
	return result

# UK GridRef to East,North
def GridRefToEN( aGridRef ):
	result	= None
	nDig	= len(aGridRef)
	if ( nDig <= 12 ) and ( nDig % 2 == 0 ):
		east	= 0
		north	= 0
		nDig	= min( 2, nDig )
		pow10	= 100000
		for i in range( nDig ):
			if ( aGridRef[i] >= "A" ) and ( aGridRef[i] <= "Z" ):
				square	= ord( aGridRef[i] ) - ord( "A" )
				if square > 7:
					square = square - 1
				eSq = floor(square % 5)
				nSq	= 4 - floor(square / 5)
				if i == 0:
					eSq = (eSq - 2)*5
					nSq = (nSq - 1)*5
				east	+= eSq*pow10
				north	+= nSq*pow10
		nDig	= max( 0, (len(aGridRef) - 2)/2 )
		pow10	= 10**( 5 - nDig )
		if nDig > 0:
			east	+= int( aGridRef[2:2+nDig], 10 )*pow10
			north	+= int( aGridRef[2+nDig:2+2*nDig], 10 )*pow10
		result	= Duple( east, north )
	return result


# Main program entry point
def main():

	DataNameT	= BIFileName
	ptsExt		= Pts1Ext
	global	Debug, Alternate, Shetland, Test, startX, startY, endX, endY

	for opt in sys.argv[1:]:
		if re.compile( "^[-/]D$", re.IGNORECASE ).match( opt ):
			Debug	= True
			print	"Debugging ..."
		elif re.compile( "^[-/]A$", re.IGNORECASE ).match( opt ):
			Alternate	= True
			ptsExt		= Pts2Ext
			print		"Alternate cell alignment method ..."
		elif re.compile( "^[-/]S$", re.IGNORECASE ).match( opt ) and not Test:
			Shetland	= True
			startX		= ShStartX
			startY		= ShStartY
			endX		= ShEndX
			endY		= ShEndY
			DataNameT	= ShFileName
			print		"Shetland only ..."
		elif re.compile( "^[-/]T$", re.IGNORECASE ).match( opt ) and not Shetland:
			Test		= True
			startX		= TstStartX
			startY		= TstStartY
			endX		= TstEndX
			endY		= TstEndY
			DataNameT	= TstFileName
			print		"Testing only ..."

	DataNameT	= os.path.join( os.path.dirname(__file__), DataNameT )
	DataOName	= os.path.join( os.path.dirname(__file__), "Output.dat" )

	for lat in range( startY, endY ):

		# Don't want an error - Test option does not require the substitution
		try:
			DataFName = DataNameT % ( lat )
		except:
			DataFName = DataNameT

		# Create a copy of the master points file with the LP & T50 heights appended
		print "Processing file: %s" % ( DataFName + ptsExt )
		Outp = open( DataOName, "w" )
		Pts	= open( DataFName + ptsExt, "r" )

		for line in Pts:
			if not re.compile( ".*,Coordinate outside transformation boundary$", re.IGNORECASE ).match( line ):
				# Parse the input line
				fields	= line.strip().split( "," )
				# Look up the Landform Panorama and Terrain 50 heights
				LPHt	= OSLP.getHeight( float(fields[6]), float(fields[7]) )
				T50Ht	= OST50.getHeight( float(fields[6]), float(fields[7]) )
				# Write all to the copy master points file
				Outp.write( ("%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%." + format("%d" % Accuracy) + "f,%." + format("%d" % Accuracy) + "f\n") % (fields[0], fields[1], fields[2], fields[3], fields[4], fields[5], fields[6], fields[7], fields[8], fields[9], LPHt, T50Ht) )

		# Close the files
		Pts.close()
		Outp.close()

		# Back up the master points file and replace with copy
		print "Backing up and replacing the points file ..."
		if os.access(DataFName + ptsExt, os.F_OK) and os.access(DataFName + bakExt, os.F_OK):
			os.remove( DataFName + bakExt )
		if os.access( DataFName + ptsExt, os.F_OK ):
			os.rename( DataFName + ptsExt, DataFName + bakExt )
		os.rename( DataOName, DataFName + ptsExt )


main()
