#!python
#
# OS2SRTM4.py
#
# Charles Macfarlane, May/June 2015
#
# Fourth of five programs/scripts to convert OS Landform Panorama or Terrain 50 elevation data to SRTM format -
# the programs/scripts are:
#	OS2SRTM1.py		Begins by making for each possible SRTM cell a WGS84 coordinate line in a *.pt? file
#					Requires GeoidEval from GeographicLib to be on the system path: http://geographiclib.sourceforge.net/
# Either:
#	GridInQuest		External OS program translates the WSG84 coordinates from above into OSGB36 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/help-and-support/navigation-technology/os-net/grid-inquest.html
# Or:
#	OS2SRTM2.js		Translates the WSG84 coordinates from above into OSGB36 and appends them
#					Unlike GIQ, this script cannot convert geiod heights, and therefore the /G switch cannot be
#					used in the final script
#	OS2SRTM3.py		Looks up heights from Landform Panorama and Terrain 50 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/products/land-form-panorama.html
#					http://www.ordnancesurvey.co.uk/business-and-government/products/terrain-50.html
#	OS2SRTM4.py		Optionally uses a bitmap OSBitMap.bin to remove points corresponding to sea
#	OS2SRTM5.py		Creates SRTM tiles
#
# This fourth script uses a bitmap of the British Isles to remove lines in the *.pt? files that correspond to
# points in the sea. The expected CSV format of *.pt? files before and after this script is run is as follows:
#	<SRTM TileName>	For example N55W006
#	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
#	<Tile Col>		Ditto
#	<Lat>			Latitude in decimal degrees corresponding to row above
#	<Lon>			Longitude in decimal degrees corresponding to col above
#	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
#	<East>			Easting in rounded metres - translated from above (Lat,Lon,Ht) by Grid In Quest
#	<North>			Northing in rounded metres - ditto
#	<EGM96 v OD>	Height in metres of the EGM96 geoid above Ordnance Datum - ditto
#	<OD Name>		Name of above Ordnance Datum
#	<LP Height>		Terrain height in metres given by OS Landform Panorama (uncorrected v EGM96)
#	<T50 Height>	Terrain height in metres given by OS Terrain 50 (uncorrected v EGM96)
# ... for example for N59W003.pt1 ...
#	N59W003,0000,0000,59.0000000,-3.0000000,52.465,342647,1012974,0.874,Newlyn (Orkney Isles only),28.183,28.441
#
# Command-line switches:
#	/D	Produce debug output
#	/A	Use alternative cell centering and correspondingly process *.pt2 file(s)
#		(*.pt1 = normal, *.pt2 = alternative - also see note 1)
#	/S	Do the six tiles covering the Shetlands only
#		(at latitude 59N, ignore corrections at or below SRTM3 row 540, SRTM1 row 1620 - also see note 2)
#	/T	Do a test run across the middle of the country
#		(see note 2)
#	/1	Create SRTM1 tiles, default is SRTM3
#		(see note 1, though here it's actually ignored as it makes no difference)
#
# Notes:
#
#	1	The /A and /1 switches MUST be used consistently throughout all the scripts.
#	2	The /S and /T switches can be introduced at any stage, but MUST be used consistently thereafter. These
#		switches are mutually exclusive, if both are given, whichever is given second is ignored.
#	3	The /G switch in the final script is only valid if Grid InQuest was used at stage 2.
#	4	Debug output is often best redirected to file, because when output is voluminous the need to scroll it
#		slows program execution.
#
# ***********************************************************************
#
# Charles Macfarlane's Work is licensed under a Creative Commons Licence -
# Attribution-Non-Commercial-No Derivative Works 2.0 UK: England & Wales
# http://creativecommons.org/licenses/by-nc-nd/2.0/uk/
#
# For further details, please see:
# http://www.macfh.co.uk/JavaJive/JavaJive.html
#
# ***********************************************************************


import os
import array
import re
import sys
import zipfile
from string import lower, upper
from math import ceil,floor,radians
from datetime import datetime

PtsPath		= ".\\"
BitMapPath	= os.path.join( "..", "..", "OS", "OSBitMap.bin" )
Pts1Ext		= ".pt1"
Pts2Ext		= ".pt2"
bakExt		= ".bak"
BIFileName	= "OS2SRTM-BI-N%d"
ShFileName	= "OS2SRTM-Shetland-N%d"
TstFileName	= "OS2SRTM-Test"
BIStartX	= -8
BIStartY	= 49
BIEndX		= 2
BIEndY		= 61
ShStartX	= -3
ShStartY	= 59
ShEndX		= 0
ShEndY		= 61
TstStartX	= -7
TstStartY	= 55
TstEndX		= 0
TstEndY		= 56
startX		= BIStartX
startY		= BIStartY
endX		= BIEndX
endY		= BIEndY

# Number of decimal places for the heights
Accuracy = 3

# Debug flag
Debug		= False
# Flags which cell alignment method to use:
#	False	Use SW corner of SRTM and OS Landform Panorama cells, adjust to middle of Terrain 50 cells
#	True	Adjust to middle of SRTM and OS Landform Panorama cells, use SW corner of Terrain 50 cells
Alternate	= False
# Shetland flag
Shetland	= False
# Test flag
Test		= False


# BitMap mask covering those areas of the BI which are mapped by OS:
# Format: 14000 x 26000 pixels, 1750 x 26000 bytes - each pixels is one bit where 1 = land, 0= sea.
class BitMapMask:

	# UK width (meters)
	nX		= 700000
	# UK height (meters)
	nY		= 1300000
	# Mask width (pixels)
	ncols	= 14000
	# Mask height (pixels)
	nrows	= 26000
	# Mask width (bytes)
	bcols	= ncols/8
	# Cell size in metres
	cell	= 50
	# BitMap Data
	data	= None

	def __init__( self ):
		self.data	= None
		# Read bitmap file
		print "Reading BitMapMask file: %s" % ( BitMapPath )
		bin = open( BitMapPath, "rb", self.nrows*self.bcols )
		self.data = array.array( "B", bin.read() )
		bin.close()

	# BitMap width
	def getCols( self ):
		return self.ncols

	# BitMap height
	def getRows( self ):
		return self.nrows

	# Cell size in metres
	def getCell( self ):
		return self.cell

	# BitMap Data
	def getData( self ):
		return self.data

	# Look up raw mask
	def getCleanBM( self, x, y ):
		bm	= 0
		if (x >= 0) and (x < self.nX) and (y >= 0) and (y < self.nY):
			x1	= x / 8
			x2	= x % 8
			bt	= 0x80 >> x2
			of	= (self.nrows - 1 - y)*self.bcols + x1
			bm	= self.data[of]
			if bm & bt != 0:
				bm = 1
			# print "/* x:%u, y:%u, of:%u, BM:%d, len:%u */<br>\n" % ( x, y, of, BM, len(tile.getData()) )
		return bm

	# Look up and interpolate bitmap
	def getBM( self, x, y ):
		cx = x / self.cell
		cy = y / self.cell
		x0 = int( floor(cx) )
		y0 = int( floor(cy) )
		x1 = x0 + 1
		y1 = y0 + 1
		dx = cx - x0;
		dy = cy - y0;

		# Possibly adjust for tile cell coordinates being of the middle, not the SW corner
		if Alternate:
			if dx < 0.5:
				if x0 > 0:	# Doubt if this will ever fail in practice
					x0 -= 1
				x1 -= 1
				dx += 0.5
			else:
				dx -= 0.5
			if dy < 0.5:
				if y0 > 0:	# Doubt if this will ever fail in practice
					y0 -= 1
				y1 -= 1
				dy += 0.5
			else:
				dy -= 0.5

		# print "/* x:%.0f, y:%.0f, cx:%.3f, cy:%.3f, x0:%u, y0:%u, x1:%u, y1:%u, dx:%.3f, dy:%.3f */<br>\n" % (x,y,cx,cy,x0,y0,x1,y1,dx,dy)
		bmSW = self.getCleanBM( x0, y0 )
		bmSE = self.getCleanBM( x1, y0 )
		bmNW = self.getCleanBM( x0, y1 )
		bmNE = self.getCleanBM( x1, y1 )
		bmS = ( 1 - dx )*bmSW + dx*bmSE
		bmN = ( 1 - dx )*bmNW + dx*bmNE
		bm = ( 1 - dy )*bmS + dy*bmN
		# print "/* x:%.0f, y:%.0f, cx:%.3f, cy:%.3f, x0:%u, y0:%u, x1:%u, y1:%u, dx:%.3f, dy:%.3f, hSW:%d, hSE:%d, hNW:%d, hNE:%d, hS:%.3f, hN:%.3f, ht:%.3f  */<br>\n" % (x,y,cx,cy,x0,y0,x1,y1,dx,dy,hSW,hSE,hNW,hNE,hS,hN,ht)
		return bm >= 0.5

# The global bitmap
BitMap = BitMapMask()


# Main program entry point
def main():

	DataNameT	= BIFileName
	ptsExt		= Pts1Ext
	global	Debug, Alternate, Shetland, Test, startX, startY, endX, endY

	for opt in sys.argv[1:]:
		if re.compile( "^[-/]D$", re.IGNORECASE ).match( opt ):
			Debug	= True
			print	"Debugging ..."
		elif re.compile( "^[-/]A$", re.IGNORECASE ).match( opt ):
			Alternate	= True
			ptsExt		= Pts2Ext
			print		"Alternate cell alignment method ..."
		elif re.compile( "^[-/]S$", re.IGNORECASE ).match( opt ) and not Test:
			Shetland	= True
			startX		= ShStartX
			startY		= ShStartY
			endX		= ShEndX
			endY		= ShEndY
			DataNameT	= ShFileName
			print		"Shetland only ..."
		elif re.compile( "^[-/]T$", re.IGNORECASE ).match( opt ) and not Shetland:
			Test		= True
			startX		= TstStartX
			startY		= TstStartY
			endX		= TstEndX
			endY		= TstEndY
			DataNameT	= TstFileName
			print		"Testing only ..."

	DataNameT	= os.path.join( PtsPath, DataNameT )
	DataOName	= os.path.join( PtsPath, "Output.dat" )

	for lat in range( startY, endY ):

		# Don't want an error - Test option does not require the substitution
		try:
			DataFName = DataNameT % ( lat )
		except:
			DataFName = DataNameT

		# Create a copy of the master points file with the LP & T50 heights appended
		print "Processing file: %s" % ( DataFName + ptsExt )
		Outp = open( DataOName, "w" )
		Pts	= open( DataFName + ptsExt, "r" )

		for line in Pts:
			if not re.compile( "Coordinate outside transformation boundary", re.IGNORECASE ).match( line ):
				# Parse the input line
				fields	= line.strip().split( "," )
				# Look up in the bitmap
				land	= BitMap.getBM( float(fields[6]), float(fields[7]) )
				if land:
					# Write all fields to the copy master points file
					Outp.write( ("%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n") % (fields[0], fields[1], fields[2], fields[3], fields[4], fields[5], fields[6], fields[7], fields[8], fields[9], fields[10], fields[11]) )

		# Close the files
		Pts.close()
		Outp.close()

		# Back up the master points file and replace with copy
		print "Backing up and replacing the points file ..."
		if os.access(DataFName + ptsExt, os.F_OK) and os.access(DataFName + bakExt, os.F_OK):
			os.remove( DataFName + bakExt )
		if os.access( DataFName + ptsExt, os.F_OK ):
			os.rename( DataFName + ptsExt, DataFName + bakExt )
		os.rename( DataOName, DataFName + ptsExt )


main()
