#!python
#
# OS2SRTM5.py
#
# Charles Macfarlane, May/June 2015
#
# Fifth and last of five programs/scripts to convert OS Landform Panorama or Terrain 50 elevation data to SRTM
# format - the programs/scripts are:
#	OS2SRTM1.py		Begins by making for each possible SRTM cell a WGS84 coordinate line in a *.pt? file
#					Requires GeoidEval from GeographicLib to be on the system path: http://geographiclib.sourceforge.net/
# Either:
#	GridInQuest		External OS program translates the WSG84 coordinates from above into OSGB36 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/help-and-support/navigation-technology/os-net/grid-inquest.html
# Or:
#	OS2SRTM2.js		Translates the WSG84 coordinates from above into OSGB36 and appends them
#					Unlike GIQ, this script cannot convert geiod heights, and therefore the /G switch cannot be
#					used in this final script
#	OS2SRTM3.py		Looks up heights from Landform Panorama and Terrain 50 and appends them
#					http://www.ordnancesurvey.co.uk/business-and-government/products/land-form-panorama.html
#					http://www.ordnancesurvey.co.uk/business-and-government/products/terrain-50.html
#	OS2SRTM4.py		Optionally uses a bitmap OSBitMap.bin to remove points corresponding to sea
#	OS2SRTM5.py		Creates SRTM tiles
#
# This fifth script takes height data created by the previous scripts in the CSV format below and uses it
# to create new SRTM tiles ...
#	<SRTM TileName>	For example N55W006
#	<Tile Row>		0-1200 (SRTM 3 arc second) or 0-3600 (SRTM 1 arc second)
#	<Tile Col>		Ditto
#	<Lat>			Latitude in decimal degrees corresponding to row above
#	<Lon>			Longitude in decimal degrees corresponding to col above
#	<EGM96 v GRS80>	Height in metres of the EGM96 geoid above the GRS80 ellipsoid at given (Lat,Lon)
#	<East>			Easting in metres - translated from above (Lat,Lon,Ht) by Grid In Quest
#	<North>			Northing in metres - ditto
#	<EGM96 v OD>	Height in metres of the EGM96 geoid above Ordnance Datum - ditto
#	<OD Name>		Name of above Ordnance Datum
#	<LP Height>		Terrain height in metres given by OS Landform Panorama (uncorrected v EGM96)
#	<T50 Height>	Terrain height in metres given by OS Terrain 50 (uncorrected v EGM96)
# ... for example for N59W003.pt1 ...
#	N59W003,0000,0000,59.0000000,-3.0000000,52.465,342647,1012974,0.874,Newlyn (Orkney Isles only),28.183,28.441
#
# Command-line switches:
#	/D	Produce debug output
#	/A	Use alternative cell centering and correspondingly process *.pt2 file(s)
#		(*.pt1 = normal, *.pt2 = alternative - also see note 1)
#	/G	Adjust for Geoid (EGM96) height (see note 3)
#	/L	Use Landform Panorama height, default is Terrain 50 height
#	/R	Create a RAW greyscale image file ".\OS*2SRTM.raw" as a visual check on the result
#		(to view this in an imaging program, note its dimensions given in the final output from this script)
#	/S	Do the six tiles covering the Shetlands only
#		(at latitude 59N, ignore corrections at or below SRTM3 row 540, SRTM1 row 1620 - also see note 2)
#	/T	Do a test run across the middle of the country
#		(see note 2)
#	/1	Create SRTM1 tiles, default is SRTM3
#		(see note 1)
#
# Notes:
#
#	1	The /A and /1 switches MUST be used consistently throughout all the scripts.
#	2	The /S and /T switches can be introduced at any stage, but MUST be used consistently thereafter. These
#		switches are mutually exclusive, if both are given, whichever is given second is ignored.
#	3	The /G switch in this final script is only valid if Grid InQuest was used at stage 2.
#	4	Debug output is often best redirected to file, because when output is voluminous the need to scroll it
#		slows program execution.
#
# ***********************************************************************
#
# Charles Macfarlane's Work is licensed under a Creative Commons Licence -
# Attribution-Non-Commercial-No Derivative Works 2.0 UK: England & Wales
# http://creativecommons.org/licenses/by-nc-nd/2.0/uk/
#
# For further details, please see:
# http://www.macfh.co.uk/JavaJive/JavaJive.html
#
# ***********************************************************************


import os
import sys
import array
import re
import zipfile
import ctypes

OutPath		= os.path.join( ".", "" )
PtsPath		= os.path.join( os.path.dirname(__file__), "" )
SRTMPath	= os.path.join( os.path.normpath( os.path.join(PtsPath, "..") ), "" )
BakExt		= ".bak"
Pts1Ext		= ".pt1"
Pts2Ext		= ".pt2"
RawExt		= ".raw"
SRTMExt		= ".hgt"
ZipExt		= ".zip"
BIFileName	= "OS2SRTM-BI-N%d"
ShFileName	= "OS2SRTM-Shetland-N%d"
TstFileName	= "OS2SRTM-Test"
BIRawName	= "OS2SRTM-BI-%s"
ShRawName	= "OS2SRTM-Shetland-%s"
TstRawName	= "OS2SRTM-Test-%s"
TileNameT	= "%s%02d%s%03d"
BIStartX	= -8
BIStartY	= 49
BIEndX		= 2
BIEndY		= 61
ShStartX	= -3
ShStartY	= 59
ShEndX		= 0
ShEndY		= 61
TstStartX	= -7
TstStartY	= 55
TstEndX		= 0
TstEndY		= 56
SRTMName	= "%s"
SRTM1Cols	= 3601
SRTM1Rows	= 3601
SRTM1Name	= SRTMName + SRTMExt
SRTM1ZName	= SRTMName + ".SRTMGL1" + SRTMExt + ZipExt
SRTM3Cols	= 1201
SRTM3Rows	= 1201
SRTM3Name	= SRTM1Name
SRTM3ZName	= SRTM3Name + ZipExt
Scale		= 255
#	Landform Panorama covers the Isle Of Man, Terrain 50 does not
#	In this tile, south of this row index is Isle Of Man
IoMTile		= "N54W005"
IoMLm		= 600
#	For Shetland only, at 59N, south of this row index is not Shetland
ShLm		= 540		#	Correctly includes Fair Isle as being part of the Shetlands
# ShLm		= 840		#	Legacy value, excluded Fair Isle

# Number of decimal places for the heights
Accuracy	= 3
# Value flagging an invalid height
InvHeight	= -32768;

# Debug flag
Debug		= False
# Flags which cell alignment method to use:
#	False	Use SW corner of SRTM and OS Landform Panorama cells, adjust to middle of Terrain 50 cells
#	True	Adjust to middle of SRTM and OS Landform Panorama cells, use SW corner of Terrain 50 cells
Alternate	= False
# Geoid flag
Geoid		= False
# Landform Panorama flag
LP			= False
# Raw flag
Raw			= False
# Shetland flag
Shetland	= False
# Test flag
Test		= False


# Main program entry point
def main():

	global Debug, Geoid, LP, Raw, Shetland, Test, IoMLm, ShLm, Scale

	startX		= BIStartX
	startY		= BIStartY
	endX		= BIEndX
	endY		= BIEndY
	DataNameT	= BIFileName
	RawFileName	= BIRawName
	ptsExt		= Pts1Ext
	SRTMCols	= SRTM3Cols
	SRTMRows	= SRTM3Rows
	SRTMName	= SRTM3Name
	SRTMZName	= SRTM3ZName

	for opt in sys.argv[1:]:
		if re.compile( "^[-/]D$", re.IGNORECASE ).match( opt ):
			Debug		= True
			print		"Debugging ..."
		elif re.compile( "^[-/]A$", re.IGNORECASE ).match( opt ):
			Alternate	= True
			ptsExt		= Pts2Ext
			print		"Alternate cell alignment method ..."
		elif re.compile( "^[-/]G$", re.IGNORECASE ).match( opt ):
			Geoid		= True
			print		"Adjusting for Geoid (EGM96) height ..."
		elif re.compile( "^[-/]L$", re.IGNORECASE ).match( opt ):
			LP			= True
			print		"Using Landform Panorama as the height source ..."
		elif re.compile( "^[-/]R$", re.IGNORECASE ).match( opt ):
			Raw			= True
			print		"RAW image will be created as visual check ..."
		elif re.compile( "^[-/]S$", re.IGNORECASE ).match( opt ) and not Test:
			Shetland	= True
			startX		= ShStartX
			startY		= ShStartY
			endX		= ShEndX
			endY		= ShEndY
			if os.access( os.path.join(PtsPath, (ShFileName + ptsExt) % startY), os.F_OK ):
				DataNameT	= ShFileName
			RawFileName	= ShRawName
			print		"Shetland only ..."
		elif re.compile( "^[-/]T$", re.IGNORECASE ).match( opt ) and not Shetland:
			Test		= True
			startX		= TstStartX
			startY		= TstStartY
			endX		= TstEndX
			endY		= TstEndY
			if os.access( os.path.join(PtsPath, TstFileName + ptsExt), os.F_OK ):
				DataNameT	= TstFileName
			RawFileName	= TstRawName
			print		"Testing only ..."
		elif re.compile( "^[-/]1$", re.IGNORECASE ).match( opt ):
			SRTMCols	= SRTM1Cols
			SRTMRows	= SRTM1Rows
			SRTMName	= SRTM1Name
			SRTMZName	= SRTM1ZName
			ShLm		*= 3
			IoMLm		*= 3
			print		"SRTM 1 ArcSecond points ..."

	DataNameT	= os.path.join( PtsPath, DataNameT + ptsExt )
	RawFileName	= RawFileName % ( "LP" if LP else "T50" )

	if Debug:
		print "\nStart Longitude: %u, Start Latitude: %u, Finish Longitude: %u, Finish Latitude: %u" % ( startX, startY, endX, endY )

	SRTMMin		= 0
	SRTMMax		= 0
	GlobTiles	= 0
	GlobChgs	= 0
	GlobVoids	= 0
	GlobFilld	= 0
	GlobMinChg	= 32767
	GlobMaxChg	= 0
	GlobAveChg	= float(0)

	print "\nScanning SRTM tiles ..."
	TileRangeX  = range ( startX, endX )
	TileRangeY  = range ( startY, endY )
	for lat in TileRangeY:
		for lon in TileRangeX:

			TileName = TileNameT % ( "N", lat, "W" if lon < 0 else "E", abs(lon) )
			print "\nCurrent SRTM tile: %s" % ( TileName )
			SRTMFName	= os.path.join( SRTMPath, SRTMName % TileName )
			SRTMZFName	= os.path.join( SRTMPath, SRTMZName % TileName )
			if os.access( SRTMFName, os.F_OK ) or os.access( SRTMZFName, os.F_OK ):

				# Init counters
				TileChgs	= 0
				TileVoids	= 0
				TileFilld	= 0
				TileMinChg	= 32767
				TileMaxChg	= 0
				TileAveChg	= float(0)

				if lat < 60:
					# Read original SRTM file
					print "Reading original SRTM file for tile: %s" % ( TileName )
					try:
						bin = open( SRTMFName, "rb", 2*SRTMCols*SRTMRows )
						SRTMTile = array.array( "h", bin.read() )
						bin.close()
					except IOError:
						try:
							bin = zipfile.ZipFile( SRTMZFName, "r" )
							SRTMTile = array.array( "h", bin.read(SRTMName % TileName) )
							bin.close()
						except IOError:
							# print "/* Could not open data file: %s */<br>\n" % fn
							pass
					if sys.byteorder == "little":
						SRTMTile.byteswap()
					# Zero noise in top row of tiles at Lat 59N
					if lat == 59:
						for x in range( SRTMCols ):
							SRTMTile[x] = 0
				else:
					print "Using blank starting tile for %s" % ( TileName )
					SRTMTile = array.array( "h", [0] * (SRTMCols * SRTMRows) )

				if Debug:
					print "SRTMTile element size: %u, memory address:%u, number of elements:%u" % (SRTMTile.itemsize, SRTMTile.buffer_info()[0], SRTMTile.buffer_info()[1] )

				# Scan for maximum and minimum heights
				for sy in range ( 0, SRTMRows ):
					for sx in range ( 0, SRTMCols ):
						so		= (SRTMRows - 1 - sy)*SRTMCols + sx
						ht		= SRTMTile[so]
						if ht != InvHeight:
							SRTMMin	= min( ht, SRTMMin )
							SRTMMax	= max( ht, SRTMMax )
						# if Debug:
						#	print "sx: %4u, sy: %4u, so: %5u, ht: %3u, min: %5u, max: %3u" % ( sx, sy, so, ht, SRTMMin, SRTMMax )

				# Don't want an error - Test option does not require the substitution
				try:
					PtsFileName = DataNameT % ( lat )
				except:
					PtsFileName = DataNameT

				# Scan points file for height corrections
				if os.access( PtsFileName, os.F_OK ):

					# Correct SRTM tile
					print "Found points file: %s" % ( PtsFileName )

					# Process points file
					pts = open( PtsFileName, "rt", 3000 )
					for line in pts:
						fields = line.strip().split(",")
						if fields[0] == TileName:
							x = int( fields[2] )
							y = int( fields[1] )
							if LP:
								new = float( fields[10] )
							else:
								new = float( fields[11] )
							if Geoid:
								gd	= float( fields[8] )
								if gd != InvHeight:
									new -= gd
							new	= int( round(new) )
							of	= (SRTMRows - 1 - y)*SRTMCols + x
							old	= SRTMTile[of]
							dif	= abs(old - new)
							if Debug:
								print "x: %4u, y: %4u, old: %4u, new: %4u, dif: % 4u" % ( x, y, old, new, dif )
							if ( dif != 0 ) and not ( not LP and (TileName == IoMTile) and (y < IoMLm) ) and not ( Shetland and (lat == 59) and (y < ShLm) ):
								SRTMTile[of] = new
								TileChgs	+= 1
								if old == InvHeight:
									TileVoids += 1
									if new != InvHeight:
										TileFilld += 1
								else:
									# if dif > 10:
									#	print "%s, lon: %0.6f, lat: %0.6f, x: %4u, y: %4u, old: %4u, new: %4u, dif: % 4u" % ( TileName, float(fields[4]), float(fields[3]), x, y, old, new, dif )
									TileAveChg 	+= dif
									TileMinChg 	= min( TileMinChg, dif )
									TileMaxChg 	= max( TileMaxChg, dif )
									SRTMMin		= min( new, SRTMMin )
									SRTMMax		= max( new, SRTMMax )
					pts.close()

					# If corrections, log them
					if TileChgs > 0:
						GlobChgs	+= TileChgs
						GlobVoids	+= TileVoids
						GlobFilld	+= TileFilld
						GlobAveChg 	+= TileAveChg
						GlobMinChg 	= min( TileMinChg, GlobMinChg )
						GlobMaxChg 	= max( TileMaxChg, GlobMaxChg )
						TileAveChg 	= TileAveChg / ( TileChgs - TileVoids )
						print "SRTM Tile %s corrections ...\nNumber of points corrected:\t%u\nData voids found:\t%u\nData voids filled:\t%u\nMinimum correction:\t%d\nMaximum correction:\t%d\nAverage correction:\t%.3f" % ( TileName, TileChgs, TileVoids, TileFilld, TileMinChg, TileMaxChg, TileAveChg )

					# If corrections or other changes, write new SRTM file
					if (lat >= 59) or (TileChgs > 0):
						GlobTiles	+= 1
						SRTMFName	= os.path.join( OutPath, SRTMName % TileName )
						bakFileName	= os.path.join( OutPath, TileName + BakExt )
						if os.access( SRTMFName, os.F_OK ) and os.access( bakFileName, os.F_OK ):
							print "Deleting existing back up file: %s" % ( bakFileName )
							os.remove( bakFileName )
						if os.access( SRTMFName, os.F_OK ):
							print "Backing up existing SRTM file: %s" % ( SRTMFName )
							os.rename( SRTMFName, bakFileName )
						print "Writing new SRTM file: %s" % ( SRTMFName )
						if sys.byteorder == "little":
							SRTMTile.byteswap()
						bin = open( SRTMFName, "wb", 2*SRTMCols*SRTMRows )
						bin.write( SRTMTile )
						bin.close()

				else:
					print "No points file found: %s!" % ( PtsFileName )

	if GlobChgs > 0:
		GlobAveChg = GlobAveChg / ( GlobChgs - GlobVoids )
	print "\nTotals ...\nNumber of tiles corrected:\t%u\nNumber of points corrected:\t%u\nData voids found:\t%u\nData voids filled:\t%u\nMinimum correction:\t%d\nMaximum correction:\t%d\nAverage correction:\t%.3f" % ( GlobTiles, GlobChgs, GlobVoids, GlobFilld, GlobMinChg, GlobMaxChg, GlobAveChg )

	# Write RAW image file for visual checking
	if Raw:
		Scale = float( Scale ) / float( SRTMMax - SRTMMin )
		print "\nMinimum height:\t%d\nMaximum height:\t%d\nScale:\t%.1f" % ( SRTMMin, SRTMMax, Scale )

		# Initialise a RAW image file for visual checking
		RawDims = ( (endX - startX) * (SRTMCols-1) + 1, (endY - startY) * (SRTMRows-1) + 1 )
		RawData = array.array( "B", chr(0) * RawDims[0] * RawDims[1] )
		if Debug:
			print "RawData element size: %u, RawData dimensions: %u x %u, number of elements:%u, memory address:%u" % ( RawData.itemsize, RawDims[0], RawDims[1], RawData.buffer_info()[1], RawData.buffer_info()[0] )

		bakFileName = os.path.join( OutPath, RawFileName + BakExt )
		RawFileName = os.path.join( OutPath, RawFileName + RawExt )
		print "\nCreating greyscale RAW image file %s as a visual check ...\n" % ( RawFileName )

		for lat in TileRangeY:
			for lon in TileRangeX:
				TileName = TileNameT % ( "N", lat, "W" if lon < 0 else "E", abs(lon) )
				print "Current SRTM tile: %s" % ( TileName )

				# Read corrected or original SRTM file and copy to RAW image
				SRTMFName	= SRTMName % ( TileName )
				SRTMTile	= None
				try:
					print "Trying to read corrected SRTM file for tile: %s" % ( TileName )
					bin = open( os.path.join(OutPath, SRTMFName), "rb", 2*SRTMCols*SRTMRows )
					SRTMTile = array.array( "h", bin.read() )
					bin.close()
				except IOError:
					try:
						print "Trying to read original SRTM file for tile: %s" % ( TileName )
						bin = open( os.path.join(SRTMPath, SRTMFName), "rb", 2*SRTMCols*SRTMRows )
						SRTMTile = array.array( "h", bin.read() )
						bin.close()
					except IOError:
						try:
							print "Trying to read original SRTM zip file for tile: %s" % ( TileName )
							bin = zipfile.ZipFile( os.path.join(SRTMPath, SRTMZName % TileName), "r" )
							SRTMTile = array.array( "h", bin.read(SRTMName % TileName) )
							bin.close()
						except IOError:
							# print "/* Could not open data file: %s */<br>\n" % fn
							pass
				if SRTMTile	!= None:
					if sys.byteorder == "little":
						SRTMTile.byteswap()
					for sy in range ( 0, SRTMRows ):
						for sx in range ( 0, SRTMCols ):
							so	= (SRTMRows - 1 - sy)*SRTMCols + sx
							ht	= SRTMTile[so]
							if ht == InvHeight:
								ht = 0
							if Debug:
								print "sx: %4u, sy: %4u, so: %5u, ht: %3u, rx: %4u, ry: %4u, ro: %5u, pi: %3u, min: %5u, max: %3u" % ( sx, sy, so, ht, rx, ry, ro, pi, SRTMMin, SRTMMax )
							rx = (lon - startX)*(SRTMCols-1) + sx
							ry = (lat - startY)*(SRTMRows-1) + sy
							ro = (RawDims[1] - 1 - ry)*RawDims[0] + rx
							pi = int( (ht - SRTMMin)*Scale )
							RawData[ro] = pi
				else:
					print "Skipping SRTM tile: %s" % ( TileName )

		print

		if os.access( bakFileName, os.F_OK ) and os.access( RawFileName, os.F_OK ):
			print "Deleting existing back up file: %s" % ( bakFileName )
			os.remove( bakFileName )
		if os.access( RawFileName, os.F_OK ):
			print "Backing up existing RAW file: %s" % ( RawFileName )
			os.rename( RawFileName, bakFileName )

		print "Writing RAW image file: %s" % ( RawFileName )
		bin = open( RawFileName, "wb", RawDims[0]*RawDims[1] )
		bin.write( RawData )
		bin.close()
		print "\nRAW image file %s created successfully!\nIts dimensions are %u x %u" % ( RawFileName, RawDims[0], RawDims[1] )


main()
